"""
Test cases for ldaptor.protocols.ldap.autofill.sambaAccount module.
"""
from twisted.trial import unittest

from ldaptor.protocols.ldap import ldapsyntax
from ldaptor.protocols.ldap.autofill import sambaAccount, sambaSamAccount
from ldaptor import testutil


class LDAPAutoFill_sambaAccount(unittest.TestCase):
    def testMustHaveObjectClass(self):
        """Test that Autofill_samba fails unless object is a sambaAccount."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["something", "other"],
            },
        )
        autoFiller = sambaAccount.Autofill_samba()
        d = o.addAutofiller(autoFiller)

        def eb(val):
            client.assertNothingSent()
            val.trap(sambaAccount.ObjectMissingObjectClassException)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testDefaultSetting(self):
        """Test that fields get their default values."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaAccount", "other"],
            },
        )
        d = o.addAutofiller(sambaAccount.Autofill_samba())

        def cb(dummy):
            client.assertNothingSent()

            self.failUnless("acctFlags" in o)
            self.failUnlessEqual(o["acctFlags"], ["[UX         ]"])

            self.failUnless("pwdLastSet" in o)
            self.failUnlessEqual(o["pwdLastSet"], ["0"])
            self.failUnless("logonTime" in o)
            self.failUnlessEqual(o["logonTime"], ["0"])
            self.failUnless("logoffTime" in o)
            self.failUnlessEqual(o["logoffTime"], ["0"])
            self.failUnless("pwdCanChange" in o)
            self.failUnlessEqual(o["pwdCanChange"], ["0"])
            self.failUnless("pwdMustChange" in o)
            self.failUnlessEqual(o["pwdMustChange"], ["0"])

        d.addCallback(cb)
        return d

    def testRid(self):
        """Test that rid field is updated based on uidNumber."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaAccount", "other"],
            },
        )
        d = o.addAutofiller(sambaAccount.Autofill_samba())

        def cb(dummy):
            client.assertNothingSent()

            o["uidNumber"] = ["1000"]
            self.failUnless("rid" in o)
            self.failUnlessEqual(o["rid"], [str(2 * 1000 + 1000)])
            o["uidNumber"] = ["1001"]
            self.failUnlessEqual(o["rid"], [str(2 * 1001 + 1000)])
            o["uidNumber"] = ["1002"]
            self.failUnlessEqual(o["rid"], [str(2 * 1002 + 1000)])
            o["uidNumber"] = ["2000"]
            self.failUnlessEqual(o["rid"], [str(2 * 2000 + 1000)])
            o["uidNumber"] = ["3000"]
            self.failUnlessEqual(o["rid"], [str(2 * 3000 + 1000)])
            o["uidNumber"] = ["0"]
            self.failUnlessEqual(o["rid"], [str(2 * 0 + 1000)])
            o["uidNumber"] = ["16000"]
            self.failUnlessEqual(o["rid"], [str(2 * 16000 + 1000)])

        d.addCallback(cb)
        return d

    def testPrimaryGroupId(self):
        """Test that primaryGroupID field is updated based on gidNumber."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaAccount", "other"],
            },
        )
        d = o.addAutofiller(sambaAccount.Autofill_samba())

        def cb(dummy):
            client.assertNothingSent()

            o["gidNumber"] = ["1000"]
            self.failUnless("primaryGroupID" in o)
            self.failUnlessEqual(o["primaryGroupID"], [str(2 * 1000 + 1001)])
            o["gidNumber"] = ["1001"]
            self.failUnlessEqual(o["primaryGroupID"], [str(2 * 1001 + 1001)])
            o["gidNumber"] = ["1002"]
            self.failUnlessEqual(o["primaryGroupID"], [str(2 * 1002 + 1001)])
            o["gidNumber"] = ["2000"]
            self.failUnlessEqual(o["primaryGroupID"], [str(2 * 2000 + 1001)])
            o["gidNumber"] = ["3000"]
            self.failUnlessEqual(o["primaryGroupID"], [str(2 * 3000 + 1001)])
            o["gidNumber"] = ["0"]
            self.failUnlessEqual(o["primaryGroupID"], [str(2 * 0 + 1001)])
            o["gidNumber"] = ["16000"]
            self.failUnlessEqual(o["primaryGroupID"], [str(2 * 16000 + 1001)])

        d.addCallback(cb)
        return d


class LDAPAutoFill_sambaSamAccount(unittest.TestCase):
    def testMustHaveObjectClass(self):
        """Test that Autofill_samba fails unless object is a sambaSamAccount."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["something", "other"],
            },
        )
        autoFiller = sambaSamAccount.Autofill_samba(domainSID="foo")
        d = o.addAutofiller(autoFiller)

        def eb(val):
            client.assertNothingSent()
            val.trap(sambaSamAccount.ObjectMissingObjectClassException)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testDefaultSetting(self):
        """Test that fields get their default values."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaSamAccount", "other"],
            },
        )
        d = o.addAutofiller(sambaSamAccount.Autofill_samba(domainSID="foo"))

        def cb(dummy):
            client.assertNothingSent()

            self.failUnlessEqual(
                set(o.keys()),
                {
                    "objectClass",
                    "sambaAcctFlags",
                    "sambaLogoffTime",
                    "sambaLogonTime",
                    "sambaPwdCanChange",
                    "sambaPwdLastSet",
                    "sambaPwdMustChange",
                },
            )

            self.failUnlessEqual(o["sambaAcctFlags"], ["[UX         ]"])
            self.failUnlessEqual(o["sambaPwdLastSet"], ["1"])
            self.failUnlessEqual(o["sambaLogonTime"], ["0"])
            self.failUnlessEqual(o["sambaLogoffTime"], ["0"])
            self.failUnlessEqual(o["sambaPwdCanChange"], ["0"])
            self.failUnlessEqual(o["sambaPwdMustChange"], ["0"])

        d.addCallback(cb)
        return d

    def testDefaultSetting_fixedPrimaryGroupSID(self):
        """Test that fields get their default values."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaSamAccount", "other"],
            },
        )
        d = o.addAutofiller(
            sambaSamAccount.Autofill_samba(
                domainSID="foo", fixedPrimaryGroupSID=4131312
            )
        )

        def cb(dummy):
            client.assertNothingSent()

            self.failUnlessEqual(
                set(o.keys()),
                {
                    "objectClass",
                    "sambaAcctFlags",
                    "sambaLogoffTime",
                    "sambaLogonTime",
                    "sambaPwdCanChange",
                    "sambaPwdLastSet",
                    "sambaPwdMustChange",
                    "sambaPrimaryGroupSID",
                },
            )

            self.failUnlessEqual(o["sambaPrimaryGroupSID"], ["foo-4131312"])
            self.failUnlessEqual(o["sambaAcctFlags"], ["[UX         ]"])
            self.failUnlessEqual(o["sambaPwdLastSet"], ["1"])
            self.failUnlessEqual(o["sambaLogonTime"], ["0"])
            self.failUnlessEqual(o["sambaLogoffTime"], ["0"])
            self.failUnlessEqual(o["sambaPwdCanChange"], ["0"])
            self.failUnlessEqual(o["sambaPwdMustChange"], ["0"])

        d.addCallback(cb)
        return d

    def testSambaSID(self):
        """Test that sambaSID field is updated based on uidNumber."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaSamAccount", "other"],
            },
        )
        d = o.addAutofiller(sambaSamAccount.Autofill_samba(domainSID="foo"))

        def cb(dummy):
            client.assertNothingSent()

            o["uidNumber"] = ["1000"]
            self.failUnless("sambaSID" in o)
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 1000 + 1000)])
            o["uidNumber"] = ["1001"]
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 1001 + 1000)])
            o["uidNumber"] = ["1002"]
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 1002 + 1000)])
            o["uidNumber"] = ["2000"]
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 2000 + 1000)])
            o["uidNumber"] = ["3000"]
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 3000 + 1000)])
            o["uidNumber"] = ["0"]
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 0 + 1000)])
            o["uidNumber"] = ["16000"]
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 16000 + 1000)])

        d.addCallback(cb)
        return d

    def testSambaSID_preExisting(self):
        """Test that sambaSID field is updated based on uidNumber."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaSamAccount", "other"],
                "uidNumber": ["1000"],
            },
        )
        d = o.addAutofiller(sambaSamAccount.Autofill_samba(domainSID="foo"))

        def cb(dummy):
            client.assertNothingSent()

            self.failUnless("sambaSID" in o)
            self.failUnlessEqual(o["sambaSID"], ["foo-%s" % (2 * 1000 + 1000)])

        d.addCallback(cb)
        return d

    def testSambaPrimaryGroupSID(self):
        """Test that sambaPrimaryGroupSID field is updated based on gidNumber."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaSamAccount", "other"],
            },
        )
        d = o.addAutofiller(sambaSamAccount.Autofill_samba(domainSID="foo"))

        def cb(dummy):
            client.assertNothingSent()

            o["gidNumber"] = ["1000"]
            self.failUnless("sambaPrimaryGroupSID" in o)
            self.failUnlessEqual(
                o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 1000 + 1001)]
            )
            o["gidNumber"] = ["1001"]
            self.failUnlessEqual(
                o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 1001 + 1001)]
            )
            o["gidNumber"] = ["1002"]
            self.failUnlessEqual(
                o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 1002 + 1001)]
            )
            o["gidNumber"] = ["2000"]
            self.failUnlessEqual(
                o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 2000 + 1001)]
            )
            o["gidNumber"] = ["3000"]
            self.failUnlessEqual(
                o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 3000 + 1001)]
            )
            o["gidNumber"] = ["0"]
            self.failUnlessEqual(o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 0 + 1001)])
            o["gidNumber"] = ["16000"]
            self.failUnlessEqual(
                o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 16000 + 1001)]
            )

        d.addCallback(cb)
        return d

    def testSambaPrimaryGroupSID_preExisting(self):
        """Test that sambaPrimaryGroupSID field is updated based on gidNumber."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaSamAccount", "other"],
                "gidNumber": ["1000"],
            },
        )
        d = o.addAutofiller(sambaSamAccount.Autofill_samba(domainSID="foo"))

        def cb(dummy):
            client.assertNothingSent()

            self.failUnless("sambaPrimaryGroupSID" in o)
            self.failUnlessEqual(
                o["sambaPrimaryGroupSID"], ["foo-%s" % (2 * 1000 + 1001)]
            )

        d.addCallback(cb)
        return d

    def testSambaPrimaryGroupSID_notUpdatedWhenFixed(self):
        """Test that sambaPrimaryGroupSID field is updated based on gidNumber."""
        client = testutil.LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["sambaSamAccount", "other"],
            },
        )
        d = o.addAutofiller(
            sambaSamAccount.Autofill_samba(domainSID="foo", fixedPrimaryGroupSID=4242)
        )

        def cb(dummy):
            client.assertNothingSent()

            self.failUnless("sambaPrimaryGroupSID" in o)
            self.failUnlessEqual(o["sambaPrimaryGroupSID"], ["foo-4242"])
            o["gidNumber"] = ["1000"]
            self.failUnless("sambaPrimaryGroupSID" in o)
            self.failUnlessEqual(o["sambaPrimaryGroupSID"], ["foo-4242"])

        d.addCallback(cb)
        return d



"""
Test cases for ldaptor.dns
"""

from twisted.trial import unittest
from ldaptor import dns


class NetmaskToNumbits(unittest.TestCase):
    def test_classA(self):
        self.assertEqual(dns.netmaskToNumbits("255.0.0.0"), 8)

    def test_classB(self):
        self.assertEqual(dns.netmaskToNumbits("255.255.0.0"), 16)

    def test_classC(self):
        self.assertEqual(dns.netmaskToNumbits("255.255.255.0"), 24)

    def test_host(self):
        self.assertEqual(dns.netmaskToNumbits("255.255.255.255"), 32)

    def test_numbits(self):
        for i in range(0, 33):
            self.assertEqual(dns.netmaskToNumbits(str(i)), i)

    def test_CIDR(self):
        for i in range(0, 33):
            mask = dns.ntoa(dns.aton(i))
            self.assertEqual(dns.netmaskToNumbits(mask), i)


class PtrSoaName(unittest.TestCase):
    def test_classA(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "255.0.0.0"), "1.in-addr.arpa.")
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "8"), "1.in-addr.arpa.")

    def test_classB(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "255.255.0.0"), "2.1.in-addr.arpa.")
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "16"), "2.1.in-addr.arpa.")

    def test_classC(self):
        self.assertEqual(
            dns.ptrSoaName("1.2.3.4", "255.255.255.0"), "3.2.1.in-addr.arpa."
        )
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "24"), "3.2.1.in-addr.arpa.")

    def test_CIDR_9(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "9"), "0/9.1.in-addr.arpa.")
        self.assertEqual(dns.ptrSoaName("1.200.3.4", "9"), "128/9.1.in-addr.arpa.")

    def test_CIDR_12(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "12"), "0/12.1.in-addr.arpa.")
        self.assertEqual(dns.ptrSoaName("1.200.3.4", "12"), "192/12.1.in-addr.arpa.")

    def test_CIDR_13(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "13"), "0/13.1.in-addr.arpa.")
        self.assertEqual(dns.ptrSoaName("1.200.3.4", "13"), "200/13.1.in-addr.arpa.")

    def test_CIDR_15(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "15"), "2/15.1.in-addr.arpa.")
        self.assertEqual(dns.ptrSoaName("1.200.3.4", "15"), "200/15.1.in-addr.arpa.")

    def test_CIDR_29(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "29"), "0/29.3.2.1.in-addr.arpa.")

    def test_CIDR_30(self):
        self.assertEqual(dns.ptrSoaName("1.2.3.4", "30"), "4/30.3.2.1.in-addr.arpa.")



import functools
from io import BytesIO

from twisted.python import failure
from twisted.internet import reactor, protocol, defer
from twisted.test import testutils


class FakeTransport(protocol.FileWrapper):
    disconnecting = False
    disconnect_done = False

    def __init__(self):
        self.data = BytesIO()
        protocol.FileWrapper.__init__(self, self.data)


class FasterIOPump(testutils.IOPump):
    def pump(self):
        """Move data back and forth.

        Returns whether any data was moved.
        """
        self.clientIO.seek(0)
        self.serverIO.seek(0)
        cData = self.clientIO.read()
        sData = self.serverIO.read()
        self.clientIO.seek(0)
        self.serverIO.seek(0)
        self.clientIO.truncate()
        self.serverIO.truncate()
        self.server.dataReceived(cData)
        self.client.dataReceived(sData)
        if cData or sData:
            return 1
        else:
            return 0


class IOPump(FasterIOPump):
    active = []

    def __init__(self, client, server, clientTransport, serverTransport):
        self.clientTransport = clientTransport
        self.serverTransport = serverTransport
        testutils.IOPump.__init__(
            self,
            client=client,
            server=server,
            clientIO=clientTransport.data,
            serverIO=serverTransport.data,
        )
        self.active.append(self)


def returnConnected(server, client):
    """Take two Protocol instances and connect them."""
    clientTransport = FakeTransport()
    client.makeConnection(clientTransport)
    serverTransport = FakeTransport()
    server.makeConnection(serverTransport)
    pump = IOPump(client, server, clientTransport, serverTransport)
    # Challenge-response authentication:
    pump.flush()
    # Uh...
    pump.flush()
    return pump


def _getDeferredResult(d):
    resultSet = []
    d.addBoth(resultSet.append)
    while not resultSet:
        for pump in IOPump.active:
            pump.pump()
        reactor.iterate()
    return resultSet[0]


def pumpingDeferredResult(d):
    result = _getDeferredResult(d)
    return result.raiseException() if isinstance(result, failure.Failure) else result


def fromCoroutineFunction(corofn):
    @functools.wraps(corofn)
    def wrapper(*args, **kwargs):
        return defer.ensureDeferred(corofn(*args, **kwargs))

    return wrapper



"""DNS-related utilities."""

from socket import inet_aton, inet_ntoa
import struct


def aton_octets(ip):
    s = inet_aton(ip)
    return struct.unpack("!I", s)[0]


def aton_numbits(num):
    n = 0
    while num > 0:
        n >>= 1
        n |= 2 ** 31
        num -= 1
    return n


def aton(ip):
    try:
        i = int(ip)
    except ValueError:
        return aton_octets(ip)
    else:
        return aton_numbits(i)


def ntoa(n):
    s = struct.pack("!I", n)
    ip = inet_ntoa(s)
    return ip


def netmaskToNumbits(netmask):
    bits = aton(netmask)
    i = 2 ** 31
    n = 0
    while bits and i > 0:
        if (bits & i) == 0:
            if bits:
                raise RuntimeError("Invalid netmask: %s" % netmask)
        n += 1
        bits -= i
        i = i >> 1
    return n


def ptrSoaName(ip, netmask):
    """
    Convert an IP address and netmask to a CIDR delegation
    -style zone name.
    """
    net = aton(ip) & aton(netmask)

    nmBits = netmaskToNumbits(netmask)
    bytes, bits = divmod(nmBits, 8)
    octets = ntoa(net).split(".")
    octets.reverse()
    if not bits:
        octets = octets[-bytes:]
    else:
        partial = octets[-bytes - 1]
        octets = octets[-bytes:]
        octets.insert(0, "%s/%d" % (partial, nmBits))

    return ".".join(octets) + ".in-addr.arpa."



"""
Test cases for ldaptor.protocols.ldap.ldapserver module.
"""
import re

import attr
from twisted.trial import unittest
from ldaptor import inmemory
from ldaptor.protocols import pureldap, pureber
from ldaptor.protocols.ldap import ldapsyntax


class TestEntryMatch(unittest.TestCase):
    def test_matchAll(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(pureldap.LDAPFilterMatchAll)
        self.assertEqual(result, True)

    def test_present_match(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(pureldap.LDAPFilter_present("aValue"))
        self.assertEqual(result, True)

    def test_present_noMatch(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(pureldap.LDAPFilter_present("noSuchValue"))
        self.assertEqual(result, False)

    def test_and_match(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_and(
                [
                    pureldap.LDAPFilter_present("aValue"),
                    pureldap.LDAPFilter_present("bValue"),
                ]
            )
        )
        self.assertEqual(result, True)

    def test_and_noMatch(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_and(
                [
                    pureldap.LDAPFilter_present("cValue"),
                    pureldap.LDAPFilter_present("dValue"),
                ]
            )
        )
        self.assertEqual(result, False)

    def test_or_match(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_or(
                [
                    pureldap.LDAPFilter_present("cValue"),
                    pureldap.LDAPFilter_present("bValue"),
                ]
            )
        )
        self.assertEqual(result, True)

    def test_or_noMatch(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_or(
                [
                    pureldap.LDAPFilter_present("cValue"),
                    pureldap.LDAPFilter_present("dValue"),
                ]
            )
        )
        self.assertEqual(result, False)

    def test_not(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_not(
                pureldap.LDAPFilter_or(
                    [
                        pureldap.LDAPFilter_present("cValue"),
                        pureldap.LDAPFilter_present("dValue"),
                    ]
                )
            )
        )
        self.assertEqual(result, True)

    def test_equality_match(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureber.BEROctetString("aValue"),
                assertionValue=pureber.BEROctetString("a"),
            )
        )
        self.assertEqual(result, True)

    def test_equality_match_caseInsensitive(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureber.BEROctetString("avaLUe"),
                assertionValue=pureber.BEROctetString("A"),
            )
        )
        self.assertEqual(result, True)

    def test_equality_noMatch(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_equalityMatch(
                attributeDesc=pureber.BEROctetString("aValue"),
                assertionValue=pureber.BEROctetString("b"),
            )
        )
        self.assertEqual(result, False)

    def test_substrings_match(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("a"),
                ],
            )
        )
        self.assertEqual(result, True)

    def test_substrings_match2(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["abcde"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("a"),
                    pureldap.LDAPFilter_substrings_final("e"),
                ],
            )
        )
        self.assertEqual(result, True)

    def test_substrings_match3(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["abcde"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("a"),
                    pureldap.LDAPFilter_substrings_any("c"),
                    pureldap.LDAPFilter_substrings_final("e"),
                ],
            )
        )
        self.assertEqual(result, True)

    def test_substrings_match4(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["abcde"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("a"),
                    pureldap.LDAPFilter_substrings_any("b"),
                    pureldap.LDAPFilter_substrings_any("c"),
                    pureldap.LDAPFilter_substrings_any("d"),
                    pureldap.LDAPFilter_substrings_final("e"),
                ],
            )
        )
        self.assertEqual(result, True)

    def test_substrings_match5(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["aoeuboeucoeudoeue"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("a"),
                    pureldap.LDAPFilter_substrings_any("b"),
                    pureldap.LDAPFilter_substrings_any("c"),
                    pureldap.LDAPFilter_substrings_any("d"),
                    pureldap.LDAPFilter_substrings_final("e"),
                ],
            )
        )
        self.assertEqual(result, True)

    def test_substrings_match6(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["aBCdE"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("A"),
                    pureldap.LDAPFilter_substrings_any("b"),
                    pureldap.LDAPFilter_substrings_any("C"),
                    pureldap.LDAPFilter_substrings_any("D"),
                    pureldap.LDAPFilter_substrings_final("e"),
                ],
            )
        )
        self.assertEqual(result, True)

    def test_substrings_match7(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["Foo"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("f"),
                ],
            )
        )
        self.assertEqual(result, True)

    def test_substrings_noMatch(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["a"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("bad"),
                    pureldap.LDAPFilter_substrings_any("dog"),
                    pureldap.LDAPFilter_substrings_any("no"),
                    pureldap.LDAPFilter_substrings_final("bone"),
                ],
            )
        )
        self.assertEqual(result, False)

    def test_substrings_noMatch2(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["aoeuboeucoeudoeue"],
                "bValue": ["b"],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_substrings(
                type="aValue",
                substrings=[
                    pureldap.LDAPFilter_substrings_initial("a"),
                    pureldap.LDAPFilter_substrings_any("b"),
                    pureldap.LDAPFilter_substrings_any("Z"),
                    pureldap.LDAPFilter_substrings_any("d"),
                    pureldap.LDAPFilter_substrings_final("e"),
                ],
            )
        )
        self.assertEqual(result, False)

    def test_greaterOrEqual_noMatch_nosuchattr(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_greaterOrEqual(
                pureber.BEROctetString("foo"), pureber.BERInteger(42)
            )
        )
        self.assertEqual(result, False)

    def test_greaterOrEqual_match_greater(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_greaterOrEqual(
                pureber.BEROctetString("num"), pureber.BERInteger(3)
            )
        )
        self.assertEqual(result, True)

    def test_greaterOrEqual_match_equal(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_greaterOrEqual(
                pureber.BEROctetString("num"), pureber.BERInteger(4)
            )
        )
        self.assertEqual(result, True)

    def test_greaterOrEqual_noMatch(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "bValue": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_greaterOrEqual(
                pureber.BEROctetString("num"), pureber.BERInteger(5)
            )
        )
        self.assertEqual(result, False)

    def test_lessOrEqual_noMatch_nosuchattr(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_lessOrEqual(
                pureber.BEROctetString("foo"), pureber.BERInteger(42)
            )
        )
        self.assertEqual(result, False)

    def test_lessOrEqual_match_less(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_lessOrEqual(
                pureber.BEROctetString("num"), pureber.BERInteger(5)
            )
        )
        self.assertEqual(result, True)

    def test_lessOrEqual_match_equal(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_lessOrEqual(
                pureber.BEROctetString("num"), pureber.BERInteger(4)
            )
        )
        self.assertEqual(result, True)

    def test_lessOrEqual_noMatch(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(
            pureldap.LDAPFilter_lessOrEqual(
                pureber.BEROctetString("num"), pureber.BERInteger(3)
            )
        )
        self.assertEqual(result, False)

    def test_extensibleMatch4(self):
        """
        An extensibleMatch filter that uses DN attributes matches an entry
        based on its OU.
        See RFC4511 section 4.5.1.
        """
        m = pureldap.LDAPFilter_extensibleMatch(
            matchingRule=None,
            type=pureldap.LDAPMatchingRuleAssertion_type(value="ou"),
            matchValue=pureldap.LDAPMatchingRuleAssertion_matchValue(value="fings"),
            dnAttributes=pureldap.LDAPMatchingRuleAssertion_dnAttributes(value=255),
        )
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,ou=fings,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(m)
        self.assertEqual(result, True)

    def test_extensibleMatch4_noMatch(self):
        """
        An extensibleMatch filter that uses DN attributes does not match an entry
        based on its OU.
        See RFC4511 section 4.5.1.
        """
        m = pureldap.LDAPFilter_extensibleMatch(
            matchingRule=None,
            type=pureldap.LDAPMatchingRuleAssertion_type(value="ou"),
            matchValue=pureldap.LDAPMatchingRuleAssertion_matchValue(value="fings"),
            dnAttributes=pureldap.LDAPMatchingRuleAssertion_dnAttributes(value=255),
        )
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,ou=uvvers,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )
        result = o.match(m)
        self.assertEqual(result, False)

    def test_notImplemented(self):
        o = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["a", "b"],
                "aValue": ["b"],
                "num": [4],
            },
        )

        @attr.s
        class UnknownMatch:
            pass

        unknownMatch = UnknownMatch()
        self.assertRaisesRegex(
            ldapsyntax.MatchNotImplemented,
            re.escape("Match type not implemented: UnknownMatch()"),
            o.match,
            unknownMatch,
        )


# TODO LDAPFilter_approxMatch
# TODO LDAPFilter_extensibleMatch



"""
Test cases for ldaptor.protocols.ldap.ldifdelta
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import ldifdelta
from ldaptor import delta, entry


class LDIFDeltaDriver(ldifdelta.LDIFDelta):
    def __init__(self):
        self.listOfCompleted = []

    def gotEntry(self, obj):
        self.listOfCompleted.append(obj)


"""
changerecord             = "changetype:" FILL
                           (change-add / change-delete /
                            change-modify / change-moddn)

change-add               = "add"                SEP 1*attrval-spec

change-delete            = "delete"             SEP

change-moddn             = ("modrdn" / "moddn") SEP
                            "newrdn:" (    FILL rdn /
                                       ":" FILL base64-rdn) SEP
                            "deleteoldrdn:" FILL ("0" / "1")  SEP
                            0*1("newsuperior:"
                            (    FILL distinguishedName /
                             ":" FILL base64-distinguishedName) SEP)

change-modify            = "modify"             SEP *mod-spec

mod-spec                 = ("add:" / "delete:" / "replace:")
                           FILL AttributeDescription SEP
                           *attrval-spec
                           "-" SEP
"""


"""
version: 1
dn: cn=foo,dc=example,dc=com
changetype: delete

"""

"""
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modrdn  #OR moddn
newrdn: rdn
deleteoldrdn: 0 #OR 1
#0..1 newsuperior: distinguishedName

"""


class TestLDIFDeltaParsing(unittest.TestCase):
    def testModification_empty(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify

"""
        )
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [
                delta.ModifyOp(dn="cn=foo,dc=example,dc=com"),
            ],
        )

    def testModification_oneAdd(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
foo: bar
-

"""
        )
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [
                delta.ModifyOp(
                    dn=b"cn=foo,dc=example,dc=com",
                    modifications=[
                        delta.Add(b"foo", [b"bar"]),
                    ],
                ),
            ],
        )

    def testModification_twoAdds(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
foo: bar
-
add: thud
thud: quux
thud: baz
-

"""
        )
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [
                delta.ModifyOp(
                    dn=b"cn=foo,dc=example,dc=com",
                    modifications=[
                        delta.Add(b"foo", [b"bar"]),
                        delta.Add(b"thud", [b"quux", b"baz"]),
                    ],
                ),
            ],
        )

    def testModification_complex(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
delete: foo
foo: bar
-
delete: garply
-
add: thud
thud: quux
thud: baz
-
replace: waldo
-
add: foo
foo: baz
-
replace: thud
thud: xyzzy
-
add: silly
-

"""
        )
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [
                delta.ModifyOp(
                    dn=b"cn=foo,dc=example,dc=com",
                    modifications=[
                        delta.Delete(b"foo", [b"bar"]),
                        delta.Delete(b"garply"),
                        delta.Add(b"thud", [b"quux", b"baz"]),
                        delta.Replace(b"waldo"),
                        delta.Add(b"foo", [b"baz"]),
                        delta.Replace(b"thud", [b"xyzzy"]),
                        delta.Add(b"silly"),
                    ],
                ),
            ],
        )

    def testModification_fail_noDash_1(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(
            ldifdelta.LDIFDeltaModificationMissingEndDashError,
            proto.dataReceived,
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
foo: bar

""",
        )

    def testModification_fail_noDash_2(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(
            ldifdelta.LDIFDeltaModificationMissingEndDashError,
            proto.dataReceived,
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo

""",
        )

    def testModification_fail_differentKey(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(
            ldifdelta.LDIFDeltaModificationDifferentAttributeTypeError,
            proto.dataReceived,
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
add: foo
bar: quux
-

""",
        )

    def testModification_fail_unknownModSpec(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(
            ldifdelta.LDIFDeltaUnknownModificationError,
            proto.dataReceived,
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: modify
fiddle: foo
foo: bar
-

""",
        )

    def testNoChangeType(self):
        """
        Raises an error is the changetype is not present.
        """
        proto = LDIFDeltaDriver()
        error = self.assertRaises(
            ldifdelta.LDIFDeltaMissingChangeTypeError,
            proto.dataReceived,
            b"""version: 1
dn: cn=foo,dc=example,dc=com
add: foo
foo: bar
-

""",
        )
        self.assertEqual((b"cn=foo,dc=example,dc=com", b"add", b"foo"), error.args)

    def testNoChangetTypeEmpty(self):
        """
        Raises an error is the changetype is not present even if no
        other command are specified for the DN.
        """
        proto = LDIFDeltaDriver()

        error = self.assertRaises(
            ldifdelta.LDIFDeltaMissingChangeTypeError,
            proto.dataReceived,
            b"""version: 1
dn: cn=foo,dc=example,dc=com

""",
        )

        self.assertEqual((b"cn=foo,dc=example,dc=com",), error.args)

    def testAdd(self):
        proto = LDIFDeltaDriver()
        proto.dataReceived(
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: add
foo: bar
thud: quux
thud: baz

"""
        )
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted,
            [
                delta.AddOp(
                    entry.BaseLDAPEntry(
                        dn=b"cn=foo,dc=example,dc=com",
                        attributes={
                            b"foo": [b"bar"],
                            b"thud": [b"quux", b"baz"],
                        },
                    )
                )
            ],
        )

    def testAdd_fail_noAttrvals(self):
        proto = LDIFDeltaDriver()
        self.assertRaises(
            ldifdelta.LDIFDeltaAddMissingAttributesError,
            proto.dataReceived,
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: add

""",
        )

    def testDelete(self):
        """ "
        Triggers a DeleteOp when the diff action is `delete`.
        """
        proto = LDIFDeltaDriver()
        proto.dataReceived(
            b"""\
version: 1
dn: cn=foo,dc=example,dc=com
changetype: delete

"""
        )
        proto.connectionLost()
        self.assertEqual(
            proto.listOfCompleted, [delta.DeleteOp(dn=b"cn=foo,dc=example,dc=com")]
        )

    def testDeleteMalformat(self):
        """ "
        Raises an error when delete change has data after the changetype.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(ldifdelta.LDIFDeltaDeleteHasJunkAfterChangeTypeError):
            proto.dataReceived(
                b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: delete
foo: bar

"""
            )

    def testMODRDN(self):
        """
        modrdn is not yet supported.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(NotImplementedError):
            proto.dataReceived(
                b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: modrdn

"""
            )

    def testMODDN(self):
        """
        moddn is not yet supported.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(NotImplementedError):
            proto.dataReceived(
                b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: moddn

"""
            )

    def testUnknownChnagetType(self):
        """
        Raises an error is the changetype is not supported.
        """
        proto = LDIFDeltaDriver()

        with self.assertRaises(ldifdelta.LDIFDeltaUnknownChangeTypeError):
            proto.dataReceived(
                b"""version: 1
dn: cn=foo,dc=example,dc=com
changetype: some-random-type

"""
            )



from ldaptor.protocols.ldap.autofill import ObjectMissingObjectClassException


class Autofill_samba:  # TODO baseclass
    def __init__(self, domainSID, fixedPrimaryGroupSID=None):
        self.domainSID = domainSID
        self.fixedPrimaryGroupSID = fixedPrimaryGroupSID

    def start(self, ldapObject):
        assert "objectClass" in ldapObject
        if "sambaSamAccount" not in ldapObject["objectClass"]:
            raise ObjectMissingObjectClassException(ldapObject)

        assert "sambaAcctFlags" not in ldapObject
        ldapObject["sambaAcctFlags"] = ["[UX         ]"]
        assert "sambaPwdLastSet" not in ldapObject
        ldapObject["sambaPwdLastSet"] = ["1"]
        assert "sambaLogonTime" not in ldapObject
        ldapObject["sambaLogonTime"] = ["0"]
        assert "sambaLogoffTime" not in ldapObject
        ldapObject["sambaLogoffTime"] = ["0"]
        assert "sambaPwdCanChange" not in ldapObject
        ldapObject["sambaPwdCanChange"] = ["0"]
        assert "sambaPwdMustChange" not in ldapObject
        ldapObject["sambaPwdMustChange"] = ["0"]

        if self.fixedPrimaryGroupSID is not None:
            assert "sambaPrimaryGroupSID" not in ldapObject
            ldapObject["sambaPrimaryGroupSID"] = [
                "%s-%d" % (self.domainSID, self.fixedPrimaryGroupSID)
            ]

        # Handle attributeTypes that were added before we got
        # started. We know we don't defer in notify, so we can do a
        # simple loop here.
        for attributeType in ldapObject.keys():
            self.notify(ldapObject, attributeType)

    def notify(self, ldapObject, attributeType):
        # sambaSID=2*uidNumber+1000
        if attributeType == "uidNumber":
            assert "uidNumber" in ldapObject
            assert len(ldapObject["uidNumber"]) == 1
            for uidNumber in ldapObject["uidNumber"]:
                uidNumber = int(uidNumber)
                sid = "%s-%d" % (self.domainSID, uidNumber * 2 + 1000)
                ldapObject["sambaSID"] = [str(sid)]
                return

        # sambaPrimaryGroupSID = fixed or 2*gidNumber+1001
        if self.fixedPrimaryGroupSID is None and attributeType == "gidNumber":
            assert "gidNumber" in ldapObject
            assert len(ldapObject["gidNumber"]) == 1
            for gidNumber in ldapObject["gidNumber"]:
                gidNumber = int(gidNumber)
                sid = "%s-%d" % (self.domainSID, gidNumber * 2 + 1001)
                ldapObject["sambaPrimaryGroupSID"] = [str(sid)]
                return



"""
Test cases for ldaptor.protocols.ldap.autofill.posixAccount module.
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import ldapsyntax, autofill
from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap.autofill import posixAccount
from ldaptor.testutil import LDAPClientTestDriver


class LDAPAutoFill_Posix(unittest.TestCase):
    def testMustHaveObjectClass(self):
        """Test that Autofill_posix fails unless object is a posixAccount."""
        client = LDAPClientTestDriver()
        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["something", "other"],
            },
        )
        autoFiller = posixAccount.Autofill_posix(baseDN="dc=example,dc=com")

        d = o.addAutofiller(autoFiller)

        failure = self.failureResultOf(d)
        self.assertIsInstance(failure.value, autofill.ObjectMissingObjectClassException)
        client.assertNothingSent()

    def testDefaultSetting(self):
        """Test that fields get their default values."""

        client = LDAPClientTestDriver(
            # uid==1000 -> free
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1000 -> taken
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="",
                    attributes=[("objectClass", ("foo", "posixAccount", "bar"))],
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1500 -> free
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1250 -> free
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1125 -> free
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1062 -> free
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1031 -> free
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="",
                    attributes=[("objectClass", ("foo", "posixAccount", "bar"))],
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1046 -> free
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1038 -> taken
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="",
                    attributes=[("objectClass", ("foo", "posixAccount", "bar"))],
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1042 -> free
            [
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1040 -> taken
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="",
                    attributes=[("objectClass", ("foo", "posixAccount", "bar"))],
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            # gid==1041 -> taken
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="",
                    attributes=[("objectClass", ("foo", "posixAccount", "bar"))],
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
        )

        o = ldapsyntax.LDAPEntryWithAutoFill(
            client=client,
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["posixAccount", "other"],
            },
        )

        d = o.addAutofiller(posixAccount.Autofill_posix(baseDN="dc=example,dc=com"))
        d.addCallback(self._cb_testDefaultSetting, client, o)
        return d

    def _cb_testDefaultSetting(self, val, client, o):
        client.assertSent(
            *[
                pureldap.LDAPSearchRequest(
                    baseObject="dc=example,dc=com",
                    scope=2,
                    derefAliases=0,
                    sizeLimit=1,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(
                            value="uidNumber"
                        ),
                        assertionValue=pureldap.LDAPAssertionValue(value="1000"),
                    ),
                    attributes=(),
                ),
            ]
            + [
                pureldap.LDAPSearchRequest(
                    baseObject="dc=example,dc=com",
                    scope=2,
                    derefAliases=0,
                    sizeLimit=1,
                    timeLimit=0,
                    typesOnly=0,
                    filter=pureldap.LDAPFilter_equalityMatch(
                        attributeDesc=pureldap.LDAPAttributeDescription(
                            value="gidNumber"
                        ),
                        assertionValue=pureldap.LDAPAssertionValue(value=str(x)),
                    ),
                    attributes=(),
                )
                for x in (
                    1000,
                    1500,
                    1250,
                    1125,
                    1062,
                    1031,
                    1046,
                    1038,
                    1042,
                    1040,
                    1041,
                )
            ]
        )

        self.failUnless("loginShell" in o)
        self.failUnlessEqual(o["loginShell"], ["/bin/sh"])

        self.failUnless("uidNumber" in o)
        self.failUnlessEqual(o["uidNumber"], ["1000"])
        self.failUnless("gidNumber" in o)
        self.failUnlessEqual(o["gidNumber"], ["1042"])



"""
An ldaptor LDAP server which can authenticate based on UPN, as AD does.

The LDAP entry needs to have the `userPrincipalName` attribute set.

dn: uid=bob,ou=people,dc=example,dc=org
objectclass: top
objectclass: person
objectClass: inetOrgPerson
uid: bob
cn: bobby
gn: Bob
sn: Roberts
mail: bob@example.org
homeDirectory: e:\\Users\\bob
userPassword: pass
userPrincipalName: bob@ad.example.org
"""

from ldaptor import interfaces
from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import distinguishedname, ldaperrors
from twisted.internet import defer
from ldaptor.protocols.ldap.ldapserver import LDAPServer


class LDAPServerWithUPNBind(LDAPServer):
    """
    An LDAP server which support BIND using UPN similar to AD.
    """

    _loginAttribute = b"userPrincipalName"

    def handle_LDAPBindRequest(self, request, controls, reply):
        if request.version != 3:
            raise ldaperrors.LDAPProtocolError(
                "Version %u not supported" % request.version
            )

        self.checkControls(controls)

        if request.dn == b"":
            # anonymous bind
            self.boundUser = None
            return pureldap.LDAPBindResponse(resultCode=0)

        root = interfaces.IConnectedLDAPEntry(self.factory)

        def _gotUPNResult(results):
            if len(results) != 1:
                # Not exactly one result, so this might not be an UNP.
                return distinguishedname.DistinguishedName(request.dn)

            # A single result, so the UPN might exist.
            return results[0].dn

        if b"@" in request.dn and b"," not in request.dn:
            # This might be an UPN request.
            filterText = b"(" + self._loginAttribute + b"=" + request.dn + b")"
            d = root.search(filterText=filterText)
            d.addCallback(_gotUPNResult)
        else:
            d = defer.succeed(distinguishedname.DistinguishedName(request.dn))

        # Once the BIND DN is known, search for the LDAP entry.
        d.addCallback(lambda dn: root.lookup(dn))

        def _noEntry(fail):
            """
            Called when the requested BIND DN was not found.
            """
            fail.trap(ldaperrors.LDAPNoSuchObject)
            return None

        d.addErrback(_noEntry)

        def _gotEntry(entry, auth):
            """
            Called when the requested BIND DN was found.
            """
            if entry is None:
                raise ldaperrors.LDAPInvalidCredentials()

            d = entry.bind(auth)

            def _cb(entry):
                """
                Called when BIND operation was successful.
                """
                self.boundUser = entry
                msg = pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.Success.resultCode,
                    matchedDN=entry.dn.getText(),
                )
                return msg

            d.addCallback(_cb)
            return d

        d.addCallback(_gotEntry, request.auth)

        return d



"""
Manage LDAP data as a tree of LDIF files.
"""
import errno
import os
import uuid

from twisted.internet import defer, error
from twisted.python import failure
from zope.interface import implementer

from ldaptor import entry, interfaces, attributeset, entryhelpers
from ldaptor.protocols.ldap import ldifprotocol, distinguishedname, ldaperrors
from ldaptor._encoder import to_unicode


class LDIFTreeEntryContainsMultipleEntries(Exception):
    """LDIFTree entry contains multiple LDIF entries."""


class LDIFTreeEntryContainsNoEntries(Exception):
    """LDIFTree entry does not contain a valid LDIF entry."""


class LDIFTreeNoSuchObject(Exception):
    """LDIFTree does not contain such entry."""


class LDAPCannotRemoveRootError(ldaperrors.LDAPNamingViolation):
    """Cannot remove root of LDAP tree"""


class StoreParsedLDIF(ldifprotocol.LDIF):
    def __init__(self):
        self.done = False
        self.seen = []

    def gotEntry(self, obj):
        self.seen.append(obj)

    def connectionLost(self, reason):
        self.done = True


def get(path, dn):
    return defer.maybeDeferred(_get, path, dn)


def _get(path, dn):
    path = to_unicode(path)
    dn = distinguishedname.DistinguishedName(dn)
    l = list(dn.split())
    assert len(l) >= 1
    l.reverse()

    parser = StoreParsedLDIF()

    entry = os.path.join(path, *["%s.dir" % rdn.getText() for rdn in l[:-1]])
    entry = os.path.join(entry, "%s.ldif" % l[-1].getText())
    f = open(entry, "rb")
    while 1:
        data = f.read(8192)
        if not data:
            break
        parser.dataReceived(data)
    parser.connectionLost(failure.Failure(error.ConnectionDone()))

    assert parser.done
    entries = parser.seen
    if len(entries) == 0:
        raise LDIFTreeEntryContainsNoEntries()
    elif len(entries) > 1:
        raise LDIFTreeEntryContainsMultipleEntries(entries)
    else:
        return entries[0]


def _putEntry(fileName, entry):
    """fileName is without extension."""
    tmp = "{}.{}.tmp".format(fileName, str(uuid.uuid4()))
    f = open(tmp, "wb")
    f.write(entry.toWire())
    f.close()
    os.rename(tmp, fileName + ".ldif")
    return True


def _put(path, entry):
    path = to_unicode(path)
    l = list(entry.dn.split())
    assert len(l) >= 1
    l.reverse()

    entryRDN = l.pop()
    if l:
        grandParent = os.path.join(path, *["%s.dir" % rdn.getText() for rdn in l[:-1]])
        parentEntry = os.path.join(grandParent, "%s.ldif" % l[-1].getText())
        parentDir = os.path.join(grandParent, "%s.dir" % l[-1].getText())
        if not os.path.exists(parentDir):
            if not os.path.exists(parentEntry):
                raise LDIFTreeNoSuchObject(entry.dn.up())
            try:
                os.mkdir(parentDir)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    # we lost a race to create the directory, safe to ignore
                    pass
                else:
                    raise
    else:
        parentDir = path
    return _putEntry(os.path.join(parentDir, "%s" % entryRDN.getText()), entry)


def put(path, entry):
    return defer.execute(_put, path, entry)


@implementer(interfaces.IConnectedLDAPEntry)
class LDIFTreeEntry(
    entry.EditableLDAPEntry,
    entryhelpers.DiffTreeMixin,
    entryhelpers.SubtreeFromChildrenMixin,
    entryhelpers.MatchMixin,
    entryhelpers.SearchByTreeWalkingMixin,
):
    def __init__(self, path, dn=None, *a, **kw):
        if dn is None:
            dn = ""
        entry.BaseLDAPEntry.__init__(self, dn, *a, **kw)
        self.path = to_unicode(path)
        if self.dn != "":
            self._load()

    def _load(self):
        assert self.path.endswith(".dir")
        entryPath = "%s.ldif" % self.path[: -len(".dir")]

        parser = StoreParsedLDIF()

        try:
            f = open(entryPath, "rb")
        except OSError as e:
            if e.errno == errno.ENOENT:
                return
            else:
                raise
        while 1:
            data = f.read(8192)
            if not data:
                break
            parser.dataReceived(data)
        parser.connectionLost(failure.Failure(error.ConnectionDone()))
        assert parser.done

        entries = parser.seen
        if len(entries) == 0:
            raise LDIFTreeEntryContainsNoEntries()
        elif len(entries) > 1:
            raise LDIFTreeEntryContainsMultipleEntries(entries)
        else:
            for k, v in entries[0].items():
                self._attributes[k] = attributeset.LDAPAttributeSet(k, v)

    def parent(self):
        if self.dn == "":
            # root
            return None
        else:
            parentPath, _ = os.path.split(self.path)
            return self.__class__(parentPath, self.dn.up())

    def _sync_children(self):
        children = []
        try:
            filenames = os.listdir(self.path)
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise
        else:
            seen = set()
            for fn in filenames:
                base, ext = os.path.splitext(fn)
                if ext not in [".dir", ".ldif"]:
                    continue
                if base in seen:
                    continue
                seen.add(base)

                dn = distinguishedname.DistinguishedName(
                    listOfRDNs=(
                        (distinguishedname.RelativeDistinguishedName(base),)
                        + self.dn.split()
                    )
                )
                e = self.__class__(os.path.join(self.path, base + ".dir"), dn)
                children.append(e)
        return children

    def _children(self, callback=None):
        children = self._sync_children()
        if callback is None:
            return children
        else:
            for c in children:
                callback(c)
            return None

    def children(self, callback=None):
        return defer.maybeDeferred(self._children, callback=callback)

    def lookup(self, dn):
        dn = distinguishedname.DistinguishedName(dn)
        if not self.dn.contains(dn):
            return defer.fail(ldaperrors.LDAPNoSuchObject(dn.getText()))
        if dn == self.dn:
            return defer.succeed(self)

        it = dn.split()
        me = self.dn.split()
        assert len(it) > len(me)
        assert (len(me) == 0) or (it[-len(me) :] == me)
        rdn = it[-len(me) - 1]
        path = os.path.join(self.path, "%s.dir" % rdn.getText())
        entry = os.path.join(self.path, "%s.ldif" % rdn.getText())
        if not os.path.isdir(path) and not os.path.isfile(entry):
            return defer.fail(ldaperrors.LDAPNoSuchObject(dn.getText()))
        else:
            childDN = distinguishedname.DistinguishedName(listOfRDNs=(rdn,) + me)
            c = self.__class__(path, childDN)
            return c.lookup(dn)

    def _addChild(self, rdn, attributes):
        rdn = distinguishedname.RelativeDistinguishedName(rdn)
        for c in self._sync_children():
            if c.dn.split()[0] == rdn:
                raise ldaperrors.LDAPEntryAlreadyExists(c.dn.getText())

        dn = distinguishedname.DistinguishedName(listOfRDNs=(rdn,) + self.dn.split())
        e = entry.BaseLDAPEntry(dn, attributes)
        if not os.path.exists(self.path):
            os.mkdir(self.path)
        fileName = os.path.join(self.path, "%s" % rdn.getText())
        tmp = "{}.{}.tmp".format(fileName, str(uuid.uuid4()))
        f = open(tmp, "wb")
        f.write(e.toWire())
        f.close()
        os.rename(tmp, fileName + ".ldif")
        dirName = os.path.join(self.path, "%s.dir" % rdn.getText())
        e = self.__class__(dirName, dn)
        return e

    def addChild(self, rdn, attributes):
        d = self._addChild(rdn, attributes)
        return d

    def _delete(self):
        if self.dn == "":
            raise LDAPCannotRemoveRootError()
        if self._sync_children():
            raise ldaperrors.LDAPNotAllowedOnNonLeaf(
                "Cannot remove entry with children: %s" % self.dn.getText()
            )
        assert self.path.endswith(".dir")
        entryPath = "%s.ldif" % self.path[: -len(".dir")]
        os.remove(entryPath)
        return self

    def delete(self):
        return defer.maybeDeferred(self._delete)

    def _deleteChild(self, rdn):
        if not isinstance(rdn, distinguishedname.RelativeDistinguishedName):
            rdn = distinguishedname.RelativeDistinguishedName(stringValue=rdn)
        for c in self._sync_children():
            if c.dn.split()[0] == rdn:
                return c.delete()
        raise ldaperrors.LDAPNoSuchObject(rdn.getText())

    def deleteChild(self, rdn):
        return defer.maybeDeferred(self._deleteChild, rdn)

    def __repr__(self):
        return "{}({!r}, {!r})".format(
            self.__class__.__name__, self.path, self.dn.getText()
        )

    def __lt__(self, other):
        if not isinstance(other, LDIFTreeEntry):
            return NotImplemented
        return self.dn < other.dn

    def __gt__(self, other):
        if not isinstance(other, LDIFTreeEntry):
            return NotImplemented
        return self.dn > other.dn

    def commit(self):
        assert self.path.endswith(".dir")
        entryPath = self.path[: -len(".dir")]
        d = defer.maybeDeferred(_putEntry, entryPath, self)

        def eb_(err):
            from twisted.python import log

            log.msg("[ERROR] Could not commit entry: {}.".format(self.dn))
            return False

        d.addErrback(eb_)
        return d

    def move(self, newDN):
        return defer.maybeDeferred(self._move, newDN)

    def _move(self, newDN):
        if not isinstance(newDN, distinguishedname.DistinguishedName):
            newDN = distinguishedname.DistinguishedName(stringValue=newDN)
        if newDN.up() != self.dn.up():
            # climb up the tree to root
            rootDN = self.dn
            rootPath = self.path
            while rootDN != "":
                rootDN = rootDN.up()
                rootPath = os.path.dirname(rootPath)
            root = self.__class__(path=rootPath, dn=rootDN)
            d = defer.maybeDeferred(root.lookup, newDN.up())
        else:
            d = defer.succeed(None)
        d.addCallback(self._move2, newDN)
        return d

    def _move2(self, newParent, newDN):
        # remove old RDN attributes
        for attr in self.dn.split()[0].split():
            self[attr.attributeType].remove(attr.value)
        # add new RDN attributes
        for attr in newDN.split()[0].split():
            self[attr.attributeType].add(attr.value)
        newRDN = newDN.split()[0]
        srcdir = os.path.dirname(self.path)
        if newParent is None:
            dstdir = srcdir
        else:
            dstdir = newParent.path

        newpath = os.path.join(dstdir, "%s.dir" % newRDN.getText())
        try:
            os.rename(self.path, newpath)
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise
        basename, ext = os.path.splitext(self.path)
        assert ext == ".dir"
        os.rename(
            "%s.ldif" % basename, os.path.join(dstdir, "%s.ldif" % newRDN.getText())
        )
        self.dn = newDN
        self.path = newpath
        return self.commit()


if __name__ == "__main__":
    """
    Demonstration LDAP server; serves an LDIFTree from given directory
    over LDAP on port 10389.
    """

    from twisted.internet import reactor, protocol
    from twisted.python import log
    import sys

    log.startLogging(sys.stderr)

    from twisted.python import components
    from ldaptor.protocols.ldap import ldapserver

    path = sys.argv[1]
    db = LDIFTreeEntry(path)

    class LDAPServerFactory(protocol.ServerFactory):
        def __init__(self, root):
            self.root = root

    class MyLDAPServer(ldapserver.LDAPServer):
        debug = True

    components.registerAdapter(
        lambda x: x.root, LDAPServerFactory, interfaces.IConnectedLDAPEntry
    )

    factory = LDAPServerFactory(db)
    factory.protocol = MyLDAPServer
    reactor.listenTCP(10389, factory)
    reactor.run()



from twisted.internet import defer
from ldaptor import delta, ldapfilter
from ldaptor._encoder import get_strings
from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import ldapsyntax, ldaperrors


def safelower(s):
    """
    As string.lower(), but return `s` if something goes wrong.
    """
    try:
        return s.lower()
    except AttributeError:
        return s


class DiffTreeMixin:
    def _diffTree_gotMyChildren(self, myChildren, other, result):
        d = other.children()
        d.addCallback(self._diffTree_gotBothChildren, myChildren, other, result)
        return d

    def _diffTree_gotBothChildren(self, otherChildren, myChildren, other, result):
        def rdnToChild(rdn, l):
            r = [x for x in l if x.dn.split()[0] == rdn]
            assert len(r) == 1
            return r[0]

        my = {x.dn.split()[0] for x in myChildren}
        his = {x.dn.split()[0] for x in otherChildren}

        # differences in common children
        commonRDN = list(my & his)
        commonRDN.sort()  # for reproducability only
        d = self._diffTree_commonChildren(
            [
                (rdnToChild(rdn, myChildren), rdnToChild(rdn, otherChildren))
                for rdn in commonRDN
            ],
            result,
        )

        # added children
        addedRDN = list(his - my)
        addedRDN.sort()  # for reproducability only
        d2 = self._diffTree_addedChildren(
            [rdnToChild(rdn, otherChildren) for rdn in addedRDN], result
        )
        d.addCallback(lambda _: d2)

        # deleted children
        deletedRDN = list(my - his)
        deletedRDN.sort()  # for reproducability only
        d3 = self._diffTree_deletedChildren(
            [rdnToChild(rdn, myChildren) for rdn in deletedRDN], result
        )
        d.addCallback(lambda _: d3)

        return d

    def _diffTree_commonChildren(self, children, result):
        if not children:
            return defer.succeed(result)
        first, rest = children[0], children[1:]
        a, b = first
        d = a.diffTree(b, result)
        d.addCallback(lambda _: self._diffTree_commonChildren(rest, result))
        return d

    def _diffTree_addedChildren(self, children, result):
        if not children:
            return result
        first, rest = children[0], children[1:]

        d = first.subtree()

        def _gotSubtree(l, result):
            for c in l:
                o = delta.AddOp(c)
                result.append(o)
            return result

        d.addCallback(_gotSubtree, result)

        d.addCallback(lambda _: self._diffTree_addedChildren(rest, result))
        return d

    def _diffTree_deletedChildren(self, children, result):
        if not children:
            return result
        first, rest = children[0], children[1:]

        d = first.subtree()

        def _gotSubtree(l, result):
            l.reverse()  # remove children before their parent
            for c in l:
                o = delta.DeleteOp(c)
                result.append(o)
            return result

        d.addCallback(_gotSubtree, result)

        d.addCallback(lambda _: self._diffTree_deletedChildren(rest, result))
        return d

    def diffTree(self, other, result=None):
        assert (
            self.dn == other.dn
        ), "diffTree arguments must refer to same LDAP tree:" "%r != %r" % (
            str(self.dn),
            str(other.dn),
        )
        if result is None:
            result = []

        # differences in root
        rootDiff = self.diff(other)
        if rootDiff is not None:
            result.append(rootDiff)

        d = self.children()
        d.addCallback(self._diffTree_gotMyChildren, other, result)

        return d


class SubtreeFromChildrenMixin:
    def subtree(self, callback=None):
        if callback is None:
            result = []
            d = self.subtree(callback=result.append)
            d.addCallback(lambda _: result)
            return d
        else:
            callback(self)
            d = self.children()

            def _gotChildren(children, callback):
                if children:
                    while len(children) > 0:
                        children.pop().subtree(callback)

            d.addCallback(_gotChildren, callback)
            return d


class MatchMixin:
    def match(self, filter):
        if isinstance(filter, pureldap.LDAPFilter_present):
            for value in get_strings(filter.value):
                if value in self:
                    return True
            return False
        elif isinstance(filter, pureldap.LDAPFilter_equalityMatch):
            # TODO case insensitivity depends on different attribute syntaxes
            for value in get_strings(filter.assertionValue.value.lower()):
                if value in [
                    val.lower() for val in self.get(filter.attributeDesc.value, [])
                ]:
                    return True
            return False
        elif isinstance(filter, pureldap.LDAPFilter_substrings):
            if filter.type not in self:
                return False
            possibleMatches = self[filter.type]
            substrings = filter.substrings[:]

            if substrings and isinstance(
                filter.substrings[0], pureldap.LDAPFilter_substrings_initial
            ):
                possibleMatches = [
                    x[len(filter.substrings[0].value) :]
                    for x in possibleMatches
                    if x.lower().startswith(filter.substrings[0].value.lower())
                ]
                del substrings[0]

            if substrings and isinstance(
                filter.substrings[-1], pureldap.LDAPFilter_substrings_final
            ):
                possibleMatches = [
                    x[: -len(filter.substrings[0].value)]
                    for x in possibleMatches
                    if x.lower().endswith(filter.substrings[-1].value.lower())
                ]
                del substrings[-1]

            while possibleMatches and substrings:
                assert isinstance(substrings[0], pureldap.LDAPFilter_substrings_any)
                r = []
                for possible in possibleMatches:
                    i = possible.lower().find(substrings[0].value.lower())
                    if i >= 0:
                        r.append(possible[i:])
                possibleMatches = r
                del substrings[0]
            if possibleMatches and not substrings:
                return True
            return False
        elif isinstance(filter, pureldap.LDAPFilter_greaterOrEqual):
            if filter.attributeDesc.value not in self:
                return False
            for value in self[filter.attributeDesc.value]:
                if value >= filter.assertionValue.value:
                    return True
            return False
        elif isinstance(filter, pureldap.LDAPFilter_lessOrEqual):
            if filter.attributeDesc.value not in self:
                return False
            for value in self[filter.attributeDesc.value]:
                if value <= filter.assertionValue.value:
                    return True
            return False
        elif isinstance(filter, pureldap.LDAPFilter_and):
            for filt in filter:
                if not self.match(filt):
                    return False
            return True
        elif isinstance(filter, pureldap.LDAPFilter_or):
            for filt in filter:
                if self.match(filt):
                    return True
            return False
        elif isinstance(filter, pureldap.LDAPFilter_not):
            return not self.match(filter.value)
        elif isinstance(filter, pureldap.LDAPFilter_extensibleMatch):
            if filter.matchingRule is None:
                attrib = filter.type.value
                match_value = filter.matchValue.value
                match_value_lower = safelower(match_value)
                if match_value_lower in [val.lower() for val in self.get(attrib, [])]:
                    return True
                for rdn in self.dn.listOfRDNs:
                    for av in rdn.attributeTypesAndValues:
                        if attrib is None or attrib == av.attributeType:
                            if match_value_lower == safelower(av.value):
                                return True
                return False
            else:
                raise ldapsyntax.MatchNotImplemented(filter)
        else:
            raise ldapsyntax.MatchNotImplemented(filter)


class SearchByTreeWalkingMixin:
    def search(
        self,
        filterText=None,
        filterObject=None,
        attributes=(),
        scope=None,
        derefAliases=None,
        sizeLimit=0,
        timeLimit=0,
        typesOnly=0,
        callback=None,
    ):
        if filterObject is None and filterText is None:
            filterObject = pureldap.LDAPFilterMatchAll
        elif filterObject is None and filterText is not None:
            filterObject = ldapfilter.parseFilter(filterText)
        elif filterObject is not None and filterText is None:
            pass
        elif filterObject is not None and filterText is not None:
            f = ldapfilter.parseFilter(filterText)
            filterObject = pureldap.LDAPFilter_and((f, filterObject))

        if scope is None:
            scope = pureldap.LDAP_SCOPE_wholeSubtree
        if derefAliases is None:
            derefAliases = pureldap.LDAP_DEREF_neverDerefAliases

        # choose iterator: base/children/subtree
        if scope == pureldap.LDAP_SCOPE_wholeSubtree:
            iterator = self.subtree
        elif scope == pureldap.LDAP_SCOPE_singleLevel:
            iterator = self.children
        elif scope == pureldap.LDAP_SCOPE_baseObject:

            def iterateSelf(callback):
                callback(self)
                return defer.succeed(None)

            iterator = iterateSelf
        else:
            raise ldaperrors.LDAPProtocolError("unknown search scope: %r" % scope)

        results = []
        if callback is None:
            matchCallback = results.append
        else:
            matchCallback = callback

        # gather results, send them
        def _tryMatch(entry):
            if entry.match(filterObject):
                matchCallback(entry)

        iterator(callback=_tryMatch)

        if callback is None:
            return defer.succeed(results)
        else:
            return defer.succeed(None)



from functools import total_ordering


from ldaptor._encoder import to_unicode, TextStrAlias

# See rfc2253
# Note that RFC 2253 sections 2.4 and 3 disagree whether "=" needs to
# be quoted. Let's trust the syntax, slapd refuses to accept unescaped
# "=" in RDN values.
escapedChars = ',+"\\<>;='
escapedChars_leading = " #"
escapedChars_trailing = " #"


def escape(s):
    r = ""
    r_trailer = ""

    if s and s[0] in escapedChars_leading:
        r = "\\" + s[0]
        s = s[1:]

    if s and s[-1] in escapedChars_trailing:
        r_trailer = "\\" + s[-1]
        s = s[:-1]

    for c in s:
        if c in escapedChars:
            r = r + "\\" + c
        elif ord(c) <= 31:
            r = r + "\\%02X" % ord(c)
        else:
            r = r + c

    return r + r_trailer


def unescape(s):
    r = ""

    while s:
        if s[0] == "\\":
            if s[1] in "0123456789abcdef":
                r = r + chr(int(s[1:3], 16))
                s = s[3:]
            else:
                r = r + s[1]
                s = s[2:]
        else:
            r = r + s[0]
            s = s[1:]

    return r


def _splitOnNotEscaped(s, separator):
    if not s:
        return []

    r = [""]
    while s:
        first = s[0:1]

        if first == "\\":
            r[-1] = r[-1] + s[:2]
            s = s[2:]
        else:

            if first == separator:
                r.append("")
                s = s[1:]
                while s[0:1] == " ":
                    s = s[1:]
            else:
                r[-1] = r[-1] + first
                s = s[1:]

    return r


class InvalidRelativeDistinguishedName(Exception):
    """
    Invalid relative distinguished name.
    It is assumed that passed RDN is of str type:
    bytes for PY2 and unicode for PY3.
    """

    def __init__(self, rdn):
        Exception.__init__(self)
        self.rdn = rdn

    def __str__(self):
        return "Invalid relative distinguished name %s." % repr(self.rdn)


class LDAPAttributeTypeAndValue(TextStrAlias):
    # TODO I should be used everywhere
    attributeType = None
    value = None

    def __init__(self, stringValue=None, attributeType=None, value=None):
        if stringValue is None:
            assert attributeType is not None
            assert value is not None
            self.attributeType = to_unicode(attributeType)
            self.value = to_unicode(value)
        else:
            assert attributeType is None
            assert value is None

            stringValue = to_unicode(stringValue)

            if "=" not in stringValue:
                raise InvalidRelativeDistinguishedName(stringValue)
            self.attributeType, self.value = stringValue.split("=", 1)

    def getText(self):
        return "=".join((escape(self.attributeType), escape(self.value)))

    def __repr__(self):
        return (
            self.__class__.__name__
            + "(attributeType="
            + repr(self.attributeType)
            + ", value="
            + repr(self.value)
            + ")"
        )

    def __hash__(self):
        return hash((self.attributeType, self.value))

    def __eq__(self, other):
        if not isinstance(other, LDAPAttributeTypeAndValue):
            return NotImplemented
        return (
            self.attributeType.lower() == other.attributeType.lower()
            and self.value.lower() == other.value.lower()
        )

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.attributeType != other.attributeType:
            return self.attributeType < other.attributeType
        else:
            return self.value < other.value

    def __gt__(self, other):
        return self != other and self > other

    def __le__(self, other):
        return not self > other

    def __ge__(self, other):
        return not self < other


class RelativeDistinguishedName(TextStrAlias):
    """LDAP Relative Distinguished Name."""

    attributeTypesAndValues = None

    def __init__(self, magic=None, stringValue=None, attributeTypesAndValues=None):
        if magic is not None:
            assert stringValue is None
            assert attributeTypesAndValues is None
            if isinstance(magic, RelativeDistinguishedName):
                attributeTypesAndValues = magic.split()
            elif isinstance(magic, (bytes, str)):
                stringValue = magic
            else:
                attributeTypesAndValues = magic

        if stringValue is None:
            assert attributeTypesAndValues is not None
            assert not isinstance(attributeTypesAndValues, (bytes, str))
            self.attributeTypesAndValues = tuple(attributeTypesAndValues)
        else:
            assert attributeTypesAndValues is None
            self.attributeTypesAndValues = tuple(
                [
                    LDAPAttributeTypeAndValue(stringValue=unescape(x))
                    for x in _splitOnNotEscaped(to_unicode(stringValue), "+")
                ]
            )

    def split(self):
        return self.attributeTypesAndValues

    def getText(self):
        return "+".join([x.getText() for x in self.attributeTypesAndValues])

    def __repr__(self):
        return (
            self.__class__.__name__
            + "(attributeTypesAndValues="
            + repr(self.attributeTypesAndValues)
            + ")"
        )

    def __hash__(self):
        return hash(self.attributeTypesAndValues)

    def __eq__(self, other):
        if not isinstance(other, RelativeDistinguishedName):
            return NotImplemented
        return self.split() == other.split()

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.split() < other.split()

    def __gt__(self, other):
        return self != other and self >= other

    def __le__(self, other):
        return not self > other

    def __ge__(self, other):
        return not self < other

    def count(self):
        return len(self.attributeTypesAndValues)


@total_ordering
class DistinguishedName(TextStrAlias):
    """LDAP Distinguished Name."""

    listOfRDNs = None

    def __init__(self, magic=None, stringValue=None, listOfRDNs=None):
        assert magic is not None or stringValue is not None or listOfRDNs is not None
        if magic is not None:
            assert stringValue is None
            assert listOfRDNs is None
            if isinstance(magic, DistinguishedName):
                listOfRDNs = magic.split()
            elif isinstance(magic, (bytes, str)):
                # This might need to be expended if we want to support
                # different encodings.
                stringValue = magic
            else:
                listOfRDNs = magic

        if stringValue is None:
            assert listOfRDNs is not None
            for x in listOfRDNs:
                assert isinstance(x, RelativeDistinguishedName)
            self.listOfRDNs = tuple(listOfRDNs)
        else:
            assert listOfRDNs is None
            self.listOfRDNs = tuple(
                [
                    RelativeDistinguishedName(stringValue=x)
                    for x in _splitOnNotEscaped(to_unicode(stringValue), ",")
                ]
            )

    def split(self):
        return self.listOfRDNs

    def up(self):
        return DistinguishedName(listOfRDNs=self.listOfRDNs[1:])

    def getText(self):
        return ",".join([x.getText() for x in self.listOfRDNs])

    def __repr__(self):
        return self.__class__.__name__ + "(listOfRDNs=" + repr(self.listOfRDNs) + ")"

    def __hash__(self):
        return hash(self.getText())

    def __eq__(self, other):
        if isinstance(other, bytes):
            return self.getText().encode("utf-8") == other
        if isinstance(other, str):
            return self.getText() == other
        if not isinstance(other, DistinguishedName):
            return NotImplemented
        return self.split() == other.split()

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        """
        Comparison used for determining the hierarchy.
        """
        if not isinstance(other, DistinguishedName):
            return NotImplemented

        # The comparison is naive and broken.
        # See https://github.com/twisted/ldaptor/issues/94
        return self.split() < other.split()

    def getDomainName(self):
        domainParts = []
        l = list(self.listOfRDNs)
        l.reverse()
        for rdn in l:
            if rdn.count() != 1:
                break
            attributeTypeAndValue = rdn.split()[0]
            if attributeTypeAndValue.attributeType.upper() != "DC":
                break
            domainParts.insert(0, attributeTypeAndValue.value)
        if domainParts:
            return ".".join(domainParts)
        else:
            return None

    def contains(self, other):
        """Does the tree rooted at DN contain or equal the other DN."""
        if self == other:
            return 1
        if not isinstance(other, DistinguishedName):
            other = DistinguishedName(other)
        its = list(other.split())
        mine = list(self.split())

        while mine and its:
            m = mine.pop()
            i = its.pop()
            if m != i:
                return 0
        if mine:
            return 0
        return 1



from ldaptor._encoder import WireStrAlias, to_bytes


def extractWord(text):
    if not text:
        return b"", b""
    l = text.split(None, 1)
    word = l[0]
    try:
        text = l[1]
    except IndexError:
        text = b""
    return word, text


def peekWord(text):
    if not text:
        return None
    return text.split(None, 1)[0]


class ASN1ParserThingie:
    def _to_list(self, text):
        """Split text into $-separated list."""
        r = []
        for x in text.split(b"$"):
            x = x.strip()
            assert x
            r.append(x)
        return tuple(r)

    def _strings_to_list(self, text):
        """Split ''-quoted strings into list."""
        r = []
        while text:
            text = text.lstrip()
            if not text:
                break
            assert text[:1] == b"'", "Text %s must start with a single quote." % repr(
                text
            )
            text = text[1:]
            end = text.index(b"'")
            r.append(text[:end])
            text = text[end + 1 :]
        return tuple(r)

    def _str_list(self, l):
        s = b" ".join([self._str(x) for x in l])
        if len(l) > 1:
            s = b"( %s )" % s
        return s

    def _list(self, l):
        s = b" $ ".join([x for x in l])
        if len(l) > 1:
            s = b"( %s )" % s
        return s

    def _str(self, s):
        return b"'%s'" % s


class ObjectClassDescription(ASN1ParserThingie, WireStrAlias):
    """
    ASN Syntax::

        d               = "0" / "1" / "2" / "3" / "4" /
                          "5" / "6" / "7" / "8" / "9"

        numericstring   = 1*d

        numericoid      = numericstring *( "." numericstring )

        space           = 1*" "

        whsp            = [ space ]

        descr           = keystring

        qdescr          = whsp "'" descr "'" whsp

        qdescrlist      = [ qdescr *( qdescr ) ]

        ; object descriptors used as schema element names
        qdescrs         = qdescr / ( whsp "(" qdescrlist ")" whsp )

        dstring         = 1*utf8

        qdstring        = whsp "'" dstring "'" whsp

        descr           = keystring

        oid             = descr / numericoid

        woid            = whsp oid whsp

        ; set of oids of either form
        oids            = woid / ( "(" oidlist ")" )

        ObjectClassDescription = "(" whsp
                numericoid whsp      ; ObjectClass identifier
                [ "NAME" qdescrs ]
                [ "DESC" qdstring ]
                [ "OBSOLETE" whsp ]
                [ "SUP" oids ]       ; Superior ObjectClasses
                [ ( "ABSTRACT" / "STRUCTURAL" / "AUXILIARY" ) whsp ]
                                     ; default structural
                [ "MUST" oids ]      ; AttributeTypes
                [ "MAY" oids ]       ; AttributeTypes
                whsp ")"
    """

    def __init__(self, text):
        self.oid = None
        self.name = None
        self.desc = None
        self.obsolete = 0
        self.sup = []
        self.type = None
        self.must = []
        self.may = []

        if text is not None:
            self._parse(to_bytes(text))

    def _parse(self, text):
        assert text[:1] == b"(", "Text %s must be in parentheses." % repr(text)
        assert text[-1:] == b")", "Text %s must be in parentheses." % repr(text)
        text = text[1:-1]
        text = text.lstrip()

        # oid
        self.oid, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"NAME":
            text = text[len(b"NAME ") :]
            text = text.lstrip()
            if text[:1] == b"'":
                text = text[1:]
                end = text.index(b"'")
                self.name = (text[:end],)
                text = text[end + 1 :]
            elif text[:1] == b"(":
                text = text[1:]
                text = text.lstrip()
                end = text.index(b")")
                self.name = self._strings_to_list(text[:end])
                text = text[end + 1 :]
            else:
                raise AssertionError()

        text = text.lstrip()

        if peekWord(text) == b"DESC":
            text = text[len(b"DESC ") :]
            text = text.lstrip()
            assert text[:1] == b"'"
            text = text[1:]
            end = text.index(b"'")
            self.desc = text[:end]
            text = text[end + 1 :]

        text = text.lstrip()

        if peekWord(text) == b"OBSOLETE":
            self.obsolete = 1
            text = text[len(b"OBSOLETE ") :]

        text = text.lstrip()

        if peekWord(text) == b"SUP":
            text = text[len(b"SUP ") :]
            text = text.lstrip()
            if text[:1] == b"(":
                text = text[1:]
                text = text.lstrip()
                end = text.index(b")")
                self.sup = self._to_list(text[:end])
                text = text[end + 1 :]
            else:
                s, text = extractWord(text)
                self.sup = [s]

        text = text.lstrip()

        if peekWord(text) == b"ABSTRACT":
            assert self.type is None
            self.type = b"ABSTRACT"
            text = text[len(b"ABSTRACT ") :]

        text = text.lstrip()

        if peekWord(text) == b"STRUCTURAL":
            assert self.type is None
            self.type = b"STRUCTURAL"
            text = text[len(b"STRUCTURAL ") :]

        text = text.lstrip()

        if peekWord(text) == b"AUXILIARY":
            assert self.type is None
            self.type = b"AUXILIARY"
            text = text[len(b"AUXILIARY ") :]

        text = text.lstrip()

        if peekWord(text) == b"MUST":
            text = text[len(b"MUST ") :]
            text = text.lstrip()
            if text[:1] == b"(":
                text = text[1:]
                text = text.lstrip()
                end = text.index(b")")
                self.must.extend(self._to_list(text[:end]))
                text = text[end + 1 :]
            else:
                s, text = extractWord(text)
                self.must.append(s)

        text = text.lstrip()

        if peekWord(text) == b"MAY":
            text = text[len(b"MAY ") :]
            text = text.lstrip()
            if text[:1] == b"(":
                text = text[1:]
                text = text.lstrip()
                end = text.index(b")")
                self.may.extend(self._to_list(text[:end]))
                text = text[end + 1 :]
            else:
                s, text = extractWord(text)
                self.may.append(s)

        text = text.lstrip()

        assert text == b"", "Text was not empty: %s" % repr(text)

        if not self.type:
            self.type = b"STRUCTURAL"

        assert self.oid
        for c in self.oid:
            assert c in b"0123456789."
        assert self.name is None or self.name
        assert self.type in (b"ABSTRACT", b"STRUCTURAL", b"AUXILIARY")

    def __repr__(self):
        nice = {}
        for k, v in self.__dict__.items():
            nice[k] = repr(v)
        return (
            "<{} instance at 0x{:x}".format(self.__class__.__name__, id(self))
            + (
                " oid=%(oid)s name=%(name)s desc=%(desc)s"
                + " obsolete=%(obsolete)s sup=%(sup)s type=%(type)s"
                + " must=%(must)s may=%(may)s>"
            )
            % nice
        )

    def toWire(self):
        r = []
        if self.name is not None:
            r.append(b"NAME %s" % self._str_list(self.name))
        if self.desc is not None:
            r.append(b"DESC %s" % self._str(self.desc))
        if self.obsolete:
            r.append(b"OBSOLETE")
        if self.sup:
            r.append(b"SUP %s" % self._list(self.sup))
        r.append(b"%s" % self.type)
        if self.must:
            r.append(b"MUST %s" % self._list(self.must))
        if self.may:
            r.append(b"MAY %s" % self._list(self.may))
        return b"( %s " % self.oid + b"\n        ".join(r) + b" )"

    def __lt__(self, other):
        if not isinstance(other, ObjectClassDescription):
            raise NotImplementedError()
        if self.name is not None and other.name is not None:
            return self.name[0].upper() < other.name[0].upper()
        else:
            return self.oid < other.oid

    def __gt__(self, other):
        if not isinstance(other, ObjectClassDescription):
            raise NotImplementedError()
        if self.name is not None and other.name is not None:
            return self.name[0].upper() > other.name[0].upper()
        else:
            return self.oid > other.oid

    def __le__(self, other):
        return self == other or self < other

    def __ge__(self, other):
        return self == other or self > other

    def __eq__(self, other):
        if not isinstance(other, ObjectClassDescription):
            raise NotImplementedError()
        return (
            self.oid == other.oid
            and self.name == other.name
            and self.desc == other.desc
            and self.obsolete == other.obsolete
            and self.sup == other.sup
            and self.type == other.type
            and self.must == other.must
            and self.may == other.may
        )

    def __ne__(self, other):
        return not (self == other)


class AttributeTypeDescription(ASN1ParserThingie, WireStrAlias):
    """
    ASN Syntax::

        AttributeTypeDescription = "(" whsp
                numericoid whsp                ; AttributeType identifier
                [ "NAME" qdescrs ]             ; name used in AttributeType
                [ "DESC" qdstring ]            ; description
                [ "OBSOLETE" whsp ]
                [ "SUP" woid ]                 ; derived from this other AttributeType
                [ "EQUALITY" woid              ; Matching Rule name
                [ "ORDERING" woid              ; Matching Rule name
                [ "SUBSTR" woid ]              ; Matching Rule name
                [ "SYNTAX" whsp noidlen whsp ] ; see section 4.3
                [ "SINGLE-VALUE" whsp ]        ; default multi-valued
                [ "COLLECTIVE" whsp ]          ; default not collective
                [ "NO-USER-MODIFICATION" whsp ]; default user modifiable
                [ "USAGE" whsp AttributeUsage ]; default userApplications
                whsp ")"

        AttributeUsage =
                "userApplications"     /
                "directoryOperation"   /
                "distributedOperation" / ; DSA-shared
                "dSAOperation"          ; DSA-specific, value depends on server

        noidlen = numericoid [ "{" len "}" ]

        len     = numericstring
    """

    def __init__(self, text):
        self.oid = None
        self.name = None
        self.desc = None
        self.obsolete = 0
        self.sup = None
        self.equality = None
        self.ordering = None
        self.substr = None
        self.syntax = None
        self.single_value = None
        self.collective = None
        self.no_user_modification = None
        self.usage = None

        # storage for experimental terms ("X-SOMETHING"), so we can
        # output them when stringifying.
        self.x_attrs = []

        if text is not None:
            self._parse(to_bytes(text))

    def _parse(self, text):
        assert text[:1] == b"(", "Text %s must be in parentheses." % repr(text)
        assert text[-1:] == b")", "Text %s must be in parentheses." % repr(text)
        text = text[1:-1]
        text = text.lstrip()

        # oid
        self.oid, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"NAME":
            text = text[len(b"NAME ") :]
            text = text.lstrip()
            if text[:1] == b"'":
                text = text[1:]
                end = text.index(b"'")
                self.name = (text[:end],)
                text = text[end + 1 :]
            elif text[:1] == b"(":
                text = text[1:]
                text = text.lstrip()
                end = text.index(b")")
                self.name = self._strings_to_list(text[:end])
                text = text[end + 1 :]
            else:
                raise AssertionError()

        text = text.lstrip()

        if peekWord(text) == b"DESC":
            text = text[len(b"DESC ") :]
            text = text.lstrip()
            assert text[:1] == b"'"
            text = text[1:]
            end = text.index(b"'")
            self.desc = text[:end]
            text = text[end + 1 :]

        text = text.lstrip()

        if peekWord(text) == b"OBSOLETE":
            self.obsolete = 1
            text = text[len(b"OBSOLETE ") :]

        text = text.lstrip()

        if peekWord(text) == b"SUP":
            text = text[len(b"SUP ") :]
            text = text.lstrip()
            self.sup, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"EQUALITY":
            text = text[len(b"EQUALITY ") :]
            text = text.lstrip()
            self.equality, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"ORDERING":
            text = text[len(b"ORDERING ") :]
            text = text.lstrip()
            self.ordering, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"SUBSTR":
            text = text[len(b"SUBSTR ") :]
            text = text.lstrip()
            self.substr, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"SYNTAX":
            text = text[len(b"SYNTAX ") :]
            text = text.lstrip()
            self.syntax, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"SINGLE-VALUE":
            assert self.single_value is None
            self.single_value = 1
            text = text[len(b"SINGLE-VALUE ") :]

        text = text.lstrip()

        if peekWord(text) == b"COLLECTIVE":
            assert self.collective is None
            self.collective = 1
            text = text[len(b"COLLECTIVE ") :]

        text = text.lstrip()

        if peekWord(text) == b"NO-USER-MODIFICATION":
            assert self.no_user_modification is None
            self.no_user_modification = 1
            text = text[len(b"NO-USER-MODIFICATION ") :]

        text = text.lstrip()

        if peekWord(text) == b"USAGE":
            assert self.usage is None
            text = text[len(b"USAGE ") :]
            text = text.lstrip()
            self.usage, text = extractWord(text)

        while True:
            text = text.lstrip()

            word = peekWord(text)
            if word is None:
                break

            if word.startswith(b"X-"):
                text = text[len(word + b" ") :]
                text = text.lstrip()
                if text[:1] == b"'":
                    text = text[1:]
                    end = text.index(b"'")
                    value = text[:end]
                    text = text[end + 1 :]
                elif text[:1] == b"(":
                    text = text[1:]
                    text = text.lstrip()
                    end = text.index(b")")
                    value = self._strings_to_list(text[:end])
                    text = text[end + 1 :]
                else:
                    raise AssertionError()

                self.x_attrs.append((word, value))
            else:
                raise AssertionError("Unhandled attributeType: %r", word)

        assert text == b"", "Text was not empty: %s" % repr(text)

        if self.single_value is None:
            self.single_value = 0

        if self.collective is None:
            self.collective = 0

        if self.no_user_modification is None:
            self.no_user_modification = 0

        assert self.oid
        for c in self.oid:
            assert c in b"0123456789."
        assert self.name is None or self.name
        assert self.usage is None or self.usage in (
            b"userApplications",
            b"directoryOperation",
            b"distributedOperation",
            b"dSAOperation",
        )

    def __repr__(self):
        nice = {}
        for k, v in self.__dict__.items():
            nice[k] = repr(v)
        return (
            "<{} instance at 0x{:x}".format(self.__class__.__name__, id(self))
            + (
                " oid=%(oid)s name=%(name)s desc=%(desc)s"
                + " obsolete=%(obsolete)s sup=%(sup)s"
                + " equality=%(equality)s ordering=%(ordering)s"
                + " substr=%(substr)s syntax=%(syntax)s"
                + " single_value=%(single_value)s"
                + " collective=%(collective)s"
                + " no_user_modification=%(no_user_modification)s"
                + " usage=%(usage)s>"
            )
            % nice
        )

    def toWire(self):
        r = []
        if self.name is not None:
            r.append(b"NAME %s" % self._str_list(self.name))
        if self.desc is not None:
            r.append(b"DESC %s" % self._str(self.desc))
        if self.obsolete:
            r.append(b"OBSOLETE")
        if self.sup is not None:
            r.append(b"SUP %s" % self.sup)
        if self.equality is not None:
            r.append(b"EQUALITY %s" % self.equality)
        if self.ordering is not None:
            r.append(b"ORDERING %s" % self.ordering)
        if self.substr is not None:
            r.append(b"SUBSTR %s" % self.substr)
        if self.syntax is not None:
            r.append(b"SYNTAX %s" % self.syntax)
        if self.single_value:
            r.append(b"SINGLE-VALUE")
        if self.collective:
            r.append(b"COLLECTIVE")
        if self.no_user_modification:
            r.append(b"NO-USER-MODIFICATION")
        if self.usage is not None:
            r.append(b"USAGE %s" % self.usage)
        for name, value in self.x_attrs:
            if isinstance(value, (bytes, str)):
                r.append(b"%s '%s'" % (name, value))
            else:
                r.append(
                    b"%s ( %s )"
                    % (
                        name,
                        b" ".join(b"'%s'" % s for s in value),
                    ),
                )
        return b"( %s " % self.oid + b"\n        ".join(r) + b" )"


class SyntaxDescription(ASN1ParserThingie, WireStrAlias):
    """
    ASN Syntax::

        SyntaxDescription = "(" whsp
                numericoid whsp
                [ "DESC" qdstring ]
                whsp ")"
    """

    def __init__(self, text):
        self.oid = None
        self.desc = None
        self.binary_transfer_required = False
        self.human_readable = True

        if text is not None:
            self._parse(to_bytes(text))

    def _parse(self, text):

        assert text[:1] == b"("
        assert text[-1:] == b")"
        text = text[1:-1]
        text = text.lstrip()

        # oid
        self.oid, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"DESC":
            text = text[len(b"DESC ") :]
            text = text.lstrip()
            assert text[:1] == b"'"
            text = text[1:]
            end = text.index(b"'")
            self.desc = text[:end]
            text = text[end + 1 :]

        text = text.lstrip()

        if peekWord(text) == b"X-BINARY-TRANSFER-REQUIRED":
            self.binary_transfer_required = True
            text = text[len(b"X-BINARY-TRANSFER-REQUIRED 'TRUE' ") :]
            text = text.lstrip()

        text = text.lstrip()

        if peekWord(text) == b"X-NOT-HUMAN-READABLE":
            self.human_readable = False
            text = text[len(b"X-NOT-HUMAN-READABLE 'TRUE' ") :]
            text = text.lstrip()

        text = text.lstrip()

        assert text == b"", "Text was not empty: %s" % repr(text)

        assert self.oid
        for c in self.oid:
            assert c in b"0123456789."

    def toWire(self):
        r = [self.oid]

        if self.desc is not None:
            r.append(b"DESC %s" % self._str(self.desc))
        if self.binary_transfer_required is True:
            r.append(b"X-BINARY-TRANSFER-REQUIRED 'TRUE'")
        if self.human_readable is False:
            r.append(b"X-NOT-HUMAN-READABLE 'TRUE'")

        return b"( " + b" ".join(r) + b" )"

    def __repr__(self):
        nice = {}
        for k, v in self.__dict__.items():
            nice[k] = repr(v)
        return (
            "<{} instance at 0x{:x}".format(self.__class__.__name__, id(self))
            + (" oid=%(oid)s desc=%(desc)s>") % nice
        )


class MatchingRuleDescription(ASN1ParserThingie, WireStrAlias):
    """
    ASN Syntax::

        MatchingRuleDescription = "(" whsp
                numericoid whsp  ; MatchingRule identifier
                [ "NAME" qdescrs ]
                [ "DESC" qdstring ]
                [ "OBSOLETE" whsp ]
                "SYNTAX" numericoid
                whsp ")"
    """

    def __init__(self, text):
        self.oid = None
        self.name = None
        self.desc = None
        self.obsolete = None
        self.syntax = None

        if text is not None:
            self._parse(to_bytes(text))

    def _parse(self, text):

        assert text[:1] == b"("
        assert text[-1:] == b")"
        text = text[1:-1]
        text = text.lstrip()

        # oid
        self.oid, text = extractWord(text)

        text = text.lstrip()

        if peekWord(text) == b"NAME":
            text = text[len(b"NAME ") :]
            text = text.lstrip()
            if text[:1] == b"'":
                text = text[1:]
                end = text.index(b"'")
                self.name = (text[:end],)
                text = text[end + 1 :]
            elif text[:1] == b"(":
                text = text[1:]
                text = text.lstrip()
                end = text.index(b")")
                self.name = self._strings_to_list(text[:end])
                text = text[end + 1 :]
            else:
                raise AssertionError()

        text = text.lstrip()

        if peekWord(text) == b"DESC":
            text = text[len(b"DESC ") :]
            text = text.lstrip()
            assert text[:1] == b"'"
            text = text[1:]
            end = text.index(b"'")
            self.desc = text[:end]
            text = text[end + 1 :]

        text = text.lstrip()

        if peekWord(text) == b"OBSOLETE":
            self.obsolete = 1
            text = text[len(b"OBSOLETE ") :]

        text = text.lstrip()

        if peekWord(text) == b"SYNTAX":
            text = text[len(b"SYNTAX ") :]
            text = text.lstrip()
            self.syntax, text = extractWord(text)

        text = text.lstrip()

        assert text == b"", "Text was not empty: %s" % repr(text)

        if self.obsolete is None:
            self.obsolete = 0
        assert self.oid
        for c in self.oid:
            assert c in b"0123456789."
        assert self.syntax

    def toWire(self):
        r = [self.oid]

        if self.name is not None:
            r.append(b"NAME %s" % self._str_list(self.name))
        if self.desc is not None:
            r.append(b"DESC %s" % self._str(self.desc))
        if self.obsolete:
            r.append(b"OBSOLETE")
        r.append(b"SYNTAX %s" % self.syntax)

        return b"( " + b" ".join(r) + b" )"

    def __repr__(self):
        nice = {}
        for k, v in self.__dict__.items():
            nice[k] = repr(v)
        return (
            "<{} instance at 0x{:x}".format(self.__class__.__name__, id(self))
            + (" oid=%(oid)s desc=%(desc)s>") % nice
        )



from twisted.internet import defer, error
from twisted.python.failure import Failure
from zope.interface import implementer

from ldaptor import interfaces, entry, entryhelpers
from ldaptor.protocols.ldap import distinguishedname, ldaperrors, ldifprotocol


class LDAPCannotRemoveRootError(ldaperrors.LDAPNamingViolation):
    """Cannot remove root of LDAP tree"""


@implementer(interfaces.IConnectedLDAPEntry)
class ReadOnlyInMemoryLDAPEntry(
    entry.EditableLDAPEntry,
    entryhelpers.DiffTreeMixin,
    entryhelpers.SubtreeFromChildrenMixin,
    entryhelpers.MatchMixin,
    entryhelpers.SearchByTreeWalkingMixin,
):
    def __init__(self, *a, **kw):
        entry.BaseLDAPEntry.__init__(self, *a, **kw)
        self._parent = None
        self._children = {}

    def parent(self):
        return self._parent

    def children(self, callback=None):
        if callback is None:
            return defer.succeed(list(self._children.values()))
        else:
            for c in self._children.values():
                callback(c)
            return defer.succeed(None)

    def _lookup(self, dn):
        if not self.dn.contains(dn):
            raise ldaperrors.LDAPNoSuchObject(dn.getText())
        if dn == self.dn:
            return defer.succeed(self)

        for c in self._children.values():
            if c.dn.contains(dn):
                return c.lookup(dn)

        raise ldaperrors.LDAPNoSuchObject(dn.getText())

    def lookup(self, dn):
        if not isinstance(dn, distinguishedname.DistinguishedName):
            dn = distinguishedname.DistinguishedName(stringValue=dn)
        return defer.maybeDeferred(self._lookup, dn)

    def fetch(self, *attributes):
        return defer.succeed(self)

    def addChild(self, rdn, attributes):
        """TODO ugly API. Returns the created entry."""
        rdn = distinguishedname.RelativeDistinguishedName(rdn)
        rdn_str = rdn.getText()
        if rdn_str in self._children:
            raise ldaperrors.LDAPEntryAlreadyExists(
                self._children[rdn_str].dn.getText()
            )
        dn = distinguishedname.DistinguishedName(listOfRDNs=(rdn,) + self.dn.split())
        e = self.__class__(dn, attributes)
        e._parent = self
        self._children[rdn_str] = e
        return e

    def _delete(self):
        if self._parent is None:
            raise LDAPCannotRemoveRootError()
        if self._children:
            raise ldaperrors.LDAPNotAllowedOnNonLeaf(self.dn)
        return self._parent.deleteChild(self.dn.split()[0])

    def delete(self):
        return defer.maybeDeferred(self._delete)

    def _deleteChild(self, rdn):
        if not isinstance(rdn, distinguishedname.RelativeDistinguishedName):
            rdn = distinguishedname.RelativeDistinguishedName(stringValue=rdn)
        rdn_str = rdn.getText()
        try:
            return self._children.pop(rdn_str)
        except KeyError:
            raise ldaperrors.LDAPNoSuchObject(rdn.getText())

    def deleteChild(self, rdn):
        return defer.maybeDeferred(self._deleteChild, rdn)

    def _move(self, newDN):
        if not isinstance(newDN, distinguishedname.DistinguishedName):
            newDN = distinguishedname.DistinguishedName(stringValue=newDN)
        if newDN.up() != self.dn.up():
            # climb up the tree to root
            root = self
            while root._parent is not None:
                root = root._parent
            d = defer.maybeDeferred(root.lookup, newDN.up())
        else:
            d = defer.succeed(None)
        d.addCallback(self._move2, newDN)
        return d

    def _move2(self, newParent, newDN):
        if newParent is not None:
            newParent._children[newDN.split()[0].getText()] = self
            del self._parent._children[self.dn.split()[0].getText()]
        # remove old RDN attributes
        for attr in self.dn.split()[0].split():
            self[attr.attributeType].remove(attr.value)
        # add new RDN attributes
        for attr in newDN.split()[0].split():
            # TODO what if the key does not exist?
            self[attr.attributeType].add(attr.value)
        self.dn = newDN
        return self

    def move(self, newDN):
        return defer.maybeDeferred(self._move, newDN)

    def commit(self):
        return defer.succeed(True)


class InMemoryLDIFProtocol(ldifprotocol.LDIF):
    """
    Receive LDIF data and gather results into an ReadOnlyInMemoryLDAPEntry.

    You can override lookupFailed and addFailed to provide smarter
    error handling. They are called as Deferred errbacks; returning
    the reason causes error to pass onward and abort the whole
    operation. Returning None from lookupFailed skips that entry, but
    continues loading.

    When the full LDIF data has been read, the completed Deferred will
    trigger.
    """

    def __init__(self):
        # Do not access this via db, just to make sure you respect the ordering
        self.db = None
        self._deferred = defer.Deferred()
        self.completed = defer.Deferred()

    def _addEntry(self, db, entry):
        d = db.lookup(entry.dn.up())
        d.addErrback(self.lookupFailed, entry)

        def _add(parent, entry):
            if parent is not None:
                parent.addChild(rdn=entry.dn.split()[0], attributes=entry)

        d.addCallback(_add, entry)
        d.addErrback(self.addFailed, entry)

        def _passDB(_, db):
            return db

        d.addCallback(_passDB, db)
        return d

    def gotEntry(self, entry):
        if self.db is None:
            # first entry, create the db, prepare to process the rest
            self.db = ReadOnlyInMemoryLDAPEntry(dn=entry.dn, attributes=entry)
            self._deferred.callback(self.db)
        else:
            self._deferred.addCallback(self._addEntry, entry)

    def lookupFailed(self, reason, entry):
        return reason  # pass the error (abort) by default

    def addFailed(self, reason, entry):
        return reason  # pass the error (abort) by default

    def connectionLost(self, reason):
        super().connectionLost(reason)
        if not reason.check(error.ConnectionDone):
            self._deferred.addCallback(lambda db: reason)
        else:
            self._deferred.chainDeferred(self.completed)

        del self._deferred  # invalidate it to flush out bugs


def fromLDIFFile(f):
    """Read LDIF data from a file."""

    p = InMemoryLDIFProtocol()
    while 1:
        data = f.read()
        if not data:
            break
        p.dataReceived(data)
    p.connectionLost(Failure(error.ConnectionDone()))

    return p.completed



# Copyright (C) 2001 Tommi Virtanen
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Test cases for ldaptor.protocols.pureber module.
"""
from twisted.trial import unittest

from ldaptor.protocols import pureber


def s(*l):
    """Join all members of list to a byte string. Integer members are converted to bytes"""
    r = b""
    for e in l:
        e = bytes((e,))
        r = r + e
    return r


def l(s):
    """Split a byte string to ord's of chars."""
    return [[x][0] for x in s]


class BerLengths(unittest.TestCase):
    knownValues = (
        (0, [0]),
        (1, [1]),
        (100, [100]),
        (126, [126]),
        (127, [127]),
        (128, [0x80 | 1, 128]),
        (129, [0x80 | 1, 129]),
        (255, [0x80 | 1, 255]),
        (256, [0x80 | 2, 1, 0]),
        (257, [0x80 | 2, 1, 1]),
        (65535, [0x80 | 2, 0xFF, 0xFF]),
        (65536, [0x80 | 3, 0x01, 0x00, 0x00]),
        (256 ** 127 - 1, [0x80 | 127] + 127 * [0xFF]),
    )

    def testToBER(self):
        for integer, encoded in self.knownValues:
            got = pureber.int2berlen(integer)
            got = bytes(got)
            got = l(got)
            self.assertEqual(got, encoded)

    def testFromBER(self):
        for integer, encoded in self.knownValues:
            m = s(*encoded)
            got, bytes = pureber.berDecodeLength(m)
            self.assertEqual(bytes, len(m))
            self.assertEqual(got, integer)

    def testPartialBER(self):
        m = bytes(pureber.int2berlen(3 * 256))
        self.assertEqual(3, len(m))
        self.assertRaises(
            pureber.BERExceptionInsufficientData, pureber.berDecodeLength, m[:2]
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData, pureber.berDecodeLength, m[:1]
        )

        m = bytes(pureber.int2berlen(256 ** 100 - 1))
        self.assertEqual(101, len(m))
        self.assertRaises(
            pureber.BERExceptionInsufficientData, pureber.berDecodeLength, m[:100]
        )


class BERBaseTests(unittest.TestCase):
    """
    Unit tests for generic BERBase.
    """

    valuesToTest = (
        (pureber.BERBase, []),
        (pureber.BERInteger, [0]),
        (pureber.BERInteger, [1]),
        (pureber.BERInteger, [4000]),
        (pureber.BERSequence, [[pureber.BERInteger(1000), pureber.BERInteger(2000)]]),
        (pureber.BERSequence, [[pureber.BERInteger(2000), pureber.BERInteger(1000)]]),
        (pureber.BEROctetString, ["foo"]),
        (pureber.BEROctetString, ["b" + chr(0xE4) + chr(0xE4)]),
    )

    def testEquality(self):
        """
        BER objects equal BER objects with same type and content
        """
        for class_, args in self.valuesToTest:
            x = class_(*args)
            y = class_(*args)
            self.assertEqual(x, x)
            self.assertEqual(x, y)

    def testInequalityWithBER(self):
        """
        BER objects do not equal BER objects with different type or content
        """
        for i in range(len(self.valuesToTest)):
            for j in range(len(self.valuesToTest)):
                if i != j:
                    i_class, i_args = self.valuesToTest[i]
                    j_class, j_args = self.valuesToTest[j]
                    x = i_class(*i_args)
                    y = j_class(*j_args)
                    self.assertNotEqual(x, y)

    def testInequalityWithNonBER(self):
        """
        BER objects are not equal with non-BER objects.
        """
        sut = pureber.BERInteger([0])

        self.assertFalse(0 == sut)
        self.assertNotEqual(0, sut)

    def testHashEquality(self):
        """
        Objects which are equal have the same hash.
        """
        for klass, arguments in self.valuesToTest:
            first = klass(*arguments)
            second = klass(*arguments)
            self.assertEqual(hash(first), hash(second))


class BERDecoderContextRepr(unittest.TestCase):
    def testRepr(self):
        self.assertEqual(
            repr(pureber.BERDecoderContext(fallback="foo", inherit="bar")),
            "<BERDecoderContext identities={"
            "0x01: BERBoolean, "
            "0x02: BERInteger, "
            "0x04: BEROctetString, "
            "0x05: BERNull, "
            "0x0a: BEREnumerated, "
            "0x10: BERSequence, "
            "0x11: BERSet"
            "} fallback='foo' inherit='bar'>",
        )


class BERIntegerKnownValues(unittest.TestCase):
    knownValues = (
        (0, [0x02, 0x01, 0]),
        (1, [0x02, 0x01, 1]),
        (2, [0x02, 0x01, 2]),
        (125, [0x02, 0x01, 125]),
        (126, [0x02, 0x01, 126]),
        (127, [0x02, 0x01, 127]),
        (-1, [0x02, 0x01, 256 - 1]),
        (-2, [0x02, 0x01, 256 - 2]),
        (-3, [0x02, 0x01, 256 - 3]),
        (-126, [0x02, 0x01, 256 - 126]),
        (-127, [0x02, 0x01, 256 - 127]),
        (-128, [0x02, 0x01, 256 - 128]),
        (-129, [0x02, 0x02, 256 - 1, 256 - 129]),
        (128, [0x02, 0x02, 0, 128]),
        (256, [0x02, 0x02, 1, 0]),
    )

    def testToBERIntegerKnownValues(self):
        """BERInteger(n).toWire() should give known result with known input"""
        for integer, encoded in self.knownValues:
            result = pureber.BERInteger(integer)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBERIntegerKnownValues(self):
        """BERInteger(encoded="...") should give known result with known input"""
        for integer, encoded in self.knownValues:
            m = s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BERInteger)
            result = result.value
            self.assertEqual(integer, result)

    def testPartialBERIntegerEncodings(self):
        """BERInteger(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BERInteger(42).toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:2],
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:1],
        )
        self.assertEqual(
            (None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), "")
        )


class BERIntegerSanityCheck(unittest.TestCase):
    def testSanity(self):
        """BERInteger(encoded=BERInteger(n)).value==n for -1000..1000"""
        for n in range(-1000, 1001, 10):
            encoded = pureber.BERInteger(n).toWire()
            result, bytes = pureber.berDecodeObject(
                pureber.BERDecoderContext(), encoded
            )
            self.assertEqual(bytes, len(encoded))
            self.assertIsInstance(result, pureber.BERInteger)
            result = result.value
            self.assertEqual(n, result)


class ObjectWithToWireMethod:
    def toWire(self):
        return b"bar"


class TestBEROctetString(unittest.TestCase):
    """
    Unit tests for BEROctetString.
    """

    knownValues = (
        ("", [0x04, 0]),
        ("foo", [0x04, 3] + l(b"foo")),
        (100 * "x", [0x04, 100] + l(100 * b"x")),
        (ObjectWithToWireMethod(), [0x04, 3] + l(b"bar")),
    )

    def testToBEROctetStringKnownValues(self):
        """BEROctetString(n).toWire() should give known result with known input"""
        for st, encoded in self.knownValues:
            result = pureber.BEROctetString(st)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBEROctetStringKnownValues(self):
        """BEROctetString(encoded="...") should give known result with known input"""
        for st, encoded in self.knownValues:
            m = s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BEROctetString)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testPartialBEROctetStringEncodings(self):
        """BEROctetString(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BEROctetString("x").toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:2],
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:1],
        )
        self.assertEqual(
            (None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), "")
        )

    def testSanity(self):
        """BEROctetString(encoded=BEROctetString(n*'x')).value==n*'x' for some values of n"""
        for n in 0, 1, 2, 3, 4, 5, 6, 100, 126, 127, 128, 129, 1000, 2000:
            encoded = pureber.BEROctetString(n * b"x").toWire()
            result, bytes = pureber.berDecodeObject(
                pureber.BERDecoderContext(), encoded
            )
            self.assertEqual(bytes, len(encoded))
            self.assertIsInstance(result, pureber.BEROctetString)
            result = result.value
            self.assertEqual(n * b"x", result)


class BERNullKnownValues(unittest.TestCase):
    def testToBERNullKnownValues(self):
        """BERNull().toWire() should give known result"""
        result = pureber.BERNull()
        result = result.toWire()
        result = l(result)
        self.assertEqual([0x05, 0x00], result)

    def testFromBERNullKnownValues(self):
        """BERNull(encoded="...") should give known result with known input"""
        encoded = [0x05, 0x00]
        m = s(*encoded)
        result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
        self.assertEqual(bytes, len(m))
        self.assertIsInstance(result, pureber.BERNull)
        self.assertEqual(0x05, result.tag)

    def testPartialBERNullEncodings(self):
        """BERNull(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BERNull().toWire()
        self.assertEqual(2, len(m))
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:1],
        )
        self.assertEqual(
            (None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), "")
        )


class BERBooleanKnownValues(unittest.TestCase):
    knownValues = (
        (0, [0x01, 0x01, 0], 0),
        (1, [0x01, 0x01, 0xFF], 0xFF),
        (2, [0x01, 0x01, 0xFF], 0xFF),
        (125, [0x01, 0x01, 0xFF], 0xFF),
        (126, [0x01, 0x01, 0xFF], 0xFF),
        (127, [0x01, 0x01, 0xFF], 0xFF),
        (-1, [0x01, 0x01, 0xFF], 0xFF),
        (-2, [0x01, 0x01, 0xFF], 0xFF),
        (-3, [0x01, 0x01, 0xFF], 0xFF),
        (-126, [0x01, 0x01, 0xFF], 0xFF),
        (-127, [0x01, 0x01, 0xFF], 0xFF),
        (-128, [0x01, 0x01, 0xFF], 0xFF),
        (-129, [0x01, 0x01, 0xFF], 0xFF),
        (-9999, [0x01, 0x01, 0xFF], 0xFF),
        (128, [0x01, 0x01, 0xFF], 0xFF),
        (255, [0x01, 0x01, 0xFF], 0xFF),
        (256, [0x01, 0x01, 0xFF], 0xFF),
        (9999, [0x01, 0x01, 0xFF], 0xFF),
    )

    def testToBERBooleanKnownValues(self):
        """BERBoolean(n).toWire() should give known result with known input"""
        for integer, encoded, dummy in self.knownValues:
            result = pureber.BERBoolean(integer)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBERBooleanKnownValues(self):
        """BERBoolean(encoded="...") should give known result with known input"""
        for integer, encoded, canon in self.knownValues:
            m = s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BERBoolean)
            result = result.value
            self.assertEqual(canon, result)

    def testPartialBERBooleanEncodings(self):
        """BERBoolean(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BERBoolean(42).toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:2],
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:1],
        )
        self.assertEqual(
            (None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), "")
        )


class BEREnumeratedKnownValues(unittest.TestCase):
    knownValues = (
        (0, [0x0A, 0x01, 0]),
        (1, [0x0A, 0x01, 1]),
        (2, [0x0A, 0x01, 2]),
        (125, [0x0A, 0x01, 125]),
        (126, [0x0A, 0x01, 126]),
        (127, [0x0A, 0x01, 127]),
        (-1, [0x0A, 0x01, 256 - 1]),
        (-2, [0x0A, 0x01, 256 - 2]),
        (-3, [0x0A, 0x01, 256 - 3]),
        (-126, [0x0A, 0x01, 256 - 126]),
        (-127, [0x0A, 0x01, 256 - 127]),
        (-128, [0x0A, 0x01, 256 - 128]),
        (-129, [0x0A, 0x02, 256 - 1, 256 - 129]),
        (128, [0x0A, 0x02, 0, 128]),
        (256, [0x0A, 0x02, 1, 0]),
    )

    def testToBEREnumeratedKnownValues(self):
        """BEREnumerated(n).toWire() should give known result with known input"""
        for integer, encoded in self.knownValues:
            result = pureber.BEREnumerated(integer)
            result = result.toWire()
            result = l(result)
            self.assertEqual(encoded, result)

    def testFromBEREnumeratedKnownValues(self):
        """BEREnumerated(encoded="...") should give known result with known input"""
        for integer, encoded in self.knownValues:
            m = s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BEREnumerated)
            result = result.value
            self.assertEqual(integer, result)

    def testPartialBEREnumeratedEncodings(self):
        """BEREnumerated(encoded="...") with too short input should throw BERExceptionInsufficientData"""
        m = pureber.BEREnumerated(42).toWire()
        self.assertEqual(3, len(m))
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:2],
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:1],
        )
        self.assertEqual(
            (None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), "")
        )


class BEREnumeratedSanityCheck(unittest.TestCase):
    def testSanity(self):
        """BEREnumerated(encoded=BEREnumerated(n)).value==n for -1000..1000"""
        for n in range(-1000, 1001, 10):
            encoded = pureber.BEREnumerated(n).toWire()
            result, bytes = pureber.berDecodeObject(
                pureber.BERDecoderContext(), encoded
            )
            self.assertEqual(bytes, len(encoded))
            self.assertIsInstance(result, pureber.BEREnumerated)
            result = result.value
            self.assertEqual(n, result)


class TestBERSequence(unittest.TestCase):
    """
    Unit test for BERSequence.
    """

    def testStringRepresentationEmpty(self):
        """
        It can return the byte string representation for empty sequence which
        is just the zero/null byte.
        """
        sut = pureber.BERSequence([])

        result = sut.toWire()

        self.assertEqual(b"0\x00", result)

    def testStringRepresentatinSmallInteger(self):
        """
        It can represent a sequence of a single integer which has a
        single byte value.
        """
        sut = pureber.BERSequence([pureber.BERInteger(2)])

        result = sut.toWire()

        self.assertEqual(b"0\x03\x02\x01\x02", result)

    def testStringRepresentatinLargerInteger(self):
        """
        It can represent a sequence of a single integer which has a
        multi bites value.
        """
        sut = pureber.BERSequence([pureber.BERInteger(128)])

        result = sut.toWire()

        self.assertEqual(b"0\x04\x02\x02\x00\x80", result)

    def testStringRepresentatinMultipleIntegers(self):
        """
        It can represent a sequence of multiple integer.
        """
        sut = pureber.BERSequence([pureber.BERInteger(3), pureber.BERInteger(128)])

        result = sut.toWire()

        self.assertEqual(b"0\x07\x02\x01\x03\x02\x02\x00\x80", result)

    def testDecodeValidInput(self):
        """
        It can be decoded from its bytes serialization.
        """
        knownValues = (
            ([], [0x30, 0x00]),
            ([pureber.BERInteger(2)], [0x30, 0x03, 0x02, 0x01, 2]),
            ([pureber.BERInteger(3)], [0x30, 0x03, 0x02, 0x01, 3]),
            ([pureber.BERInteger(128)], [0x30, 0x04, 0x02, 0x02, 0, 128]),
            (
                [
                    pureber.BERInteger(2),
                    pureber.BERInteger(3),
                    pureber.BERInteger(128),
                ],
                [0x30, 0x0A] + [0x02, 0x01, 2] + [0x02, 0x01, 3] + [0x02, 0x02, 0, 128],
            ),
        )

        for content, encoded in knownValues:
            m = s(*encoded)
            result, bytes = pureber.berDecodeObject(pureber.BERDecoderContext(), m)
            self.assertEqual(bytes, len(m))
            self.assertIsInstance(result, pureber.BERSequence)
            result = result.data
            self.assertEqual(len(content), len(result))
            for i in range(len(content)):
                self.assertEqual(content[i], result[i])
            self.assertEqual(content, result)

    def testDecdeInvalidInput(self):
        """
        It raises BERExceptionInsufficientData when trying to decode from
        data which is not valid.
        """
        m = pureber.BERSequence([pureber.BERInteger(2)]).toWire()
        self.assertEqual(5, len(m))

        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:4],
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:3],
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:2],
        )
        self.assertRaises(
            pureber.BERExceptionInsufficientData,
            pureber.berDecodeObject,
            pureber.BERDecoderContext(),
            m[:1],
        )
        self.assertEqual(
            (None, 0), pureber.berDecodeObject(pureber.BERDecoderContext(), "")
        )



#!/usr/bin/env python3


from ldaptor.protocols import pureldap
from ldaptor._encoder import to_unicode


"""

RFC2254:

        filter     = "(" filtercomp ")"
        filtercomp = and / or / not / item
        and        = "&" filterlist
        or         = "|" filterlist
        not        = "!" filter
        filterlist = 1*filter
        item       = simple / present / substring / extensible
        simple     = attr filtertype value
        filtertype = equal / approx / greater / less
        equal      = "="
        approx     = "~="
        greater    = ">="
        less       = "<="
        extensible = attr [":dn"] [":" matchingrule] ":=" value
                     / [":dn"] ":" matchingrule ":=" value
        present    = attr "=*"
        substring  = attr "=" [initial] any [final]
        initial    = value
        any        = "*" *(value "*")
        final      = value
        attr       = AttributeDescription from Section 4.1.5 of [1]
        matchingrule = MatchingRuleId from Section 4.1.9 of [1]
        value      = AttributeValue from Section 4.1.6 of [1]
"""


class InvalidLDAPFilter(Exception):
    def __init__(self, msg, loc, text):
        Exception.__init__(self)
        self.msg = msg
        self.loc = loc
        self.text = text

    def __str__(self):
        return "Invalid LDAP filter: %s at point %d in %r" % (
            self.msg,
            self.loc,
            self.text,
        )


def parseExtensible(attr, s):
    raise NotImplementedError()


from pyparsing import (
    Word,
    Literal,
    Optional,
    ZeroOrMore,
    Suppress,
    Group,
    Forward,
    OneOrMore,
    ParseException,
    CharsNotIn,
    Combine,
    StringStart,
    StringEnd,
    delimitedList,
)

import string

filter_ = Forward()
attr = Word(
    string.ascii_letters,
    string.ascii_letters + string.digits + ";-",
)
attr.leaveWhitespace()
attr.setName("attr")
hexdigits = Word(string.hexdigits, exact=2)
hexdigits.setName("hexdigits")
escaped = Suppress(Literal("\\")) + hexdigits
escaped.setName("escaped")


def _p_escaped(s, l, t):
    text = t[0]
    return chr(int(text, 16))


escaped.setParseAction(_p_escaped)
value = Combine(OneOrMore(CharsNotIn("*()\\\0") | escaped))
value.setName("value")
equal = Literal("=")
equal.setParseAction(lambda s, l, t: pureldap.LDAPFilter_equalityMatch)
approx = Literal("~=")
approx.setParseAction(lambda s, l, t: pureldap.LDAPFilter_approxMatch)
greater = Literal(">=")
greater.setParseAction(lambda s, l, t: pureldap.LDAPFilter_greaterOrEqual)
less = Literal("<=")
less.setParseAction(lambda s, l, t: pureldap.LDAPFilter_lessOrEqual)
filtertype = equal | approx | greater | less
filtertype.setName("filtertype")
simple = attr + filtertype + value
simple.leaveWhitespace()
simple.setName("simple")


def _p_simple(s, l, t):
    attr, filtertype, value = t
    return filtertype(
        attributeDesc=pureldap.LDAPAttributeDescription(attr),
        assertionValue=pureldap.LDAPAssertionValue(value),
    )


simple.setParseAction(_p_simple)
present = attr + "=*"
present.setParseAction(lambda s, l, t: pureldap.LDAPFilter_present(t[0]))
initial = value.copy()
initial.setParseAction(lambda s, l, t: pureldap.LDAPFilter_substrings_initial(t[0]))
initial.setName("initial")
any_value = value + Suppress(Literal("*"))
any_value.setParseAction(lambda s, l, t: pureldap.LDAPFilter_substrings_any(t[0]))
any = Suppress(Literal("*")) + ZeroOrMore(any_value)
any.setName("any")
final = value.copy()
final.setName("final")
final.setParseAction(lambda s, l, t: pureldap.LDAPFilter_substrings_final(t[0]))
substring = (
    attr + Suppress(Literal("=")) + Group(Optional(initial) + any + Optional(final))
)
substring.setName("substring")


def _p_substring(s, l, t):
    attrtype, substrings = t
    return pureldap.LDAPFilter_substrings(type=attrtype, substrings=substrings)


substring.setParseAction(_p_substring)

keystring = Word(string.ascii_letters, string.ascii_letters + string.digits + ";-")
keystring.setName("keystring")
numericoid = delimitedList(Word(string.digits), delim=".", combine=True)
numericoid.setName("numericoid")
oid = numericoid | keystring
oid.setName("oid")
matchingrule = oid.copy()
matchingrule.setName("matchingrule")

extensible_dn = Optional(":dn")


def _p_extensible_dn(s, l, t):
    return bool(t)


extensible_dn.setParseAction(_p_extensible_dn)

matchingrule_or_none = Optional(Suppress(":") + matchingrule)


def _p_matchingrule_or_none(s, l, t):
    if not t:
        return [None]
    else:
        return t[0]


matchingrule_or_none.setParseAction(_p_matchingrule_or_none)

extensible_attr = attr + extensible_dn + matchingrule_or_none + Suppress(":=") + value
extensible_attr.setName("extensible_attr")


def _p_extensible_attr(s, l, t):
    return list(t)


extensible_attr.setParseAction(_p_extensible_attr)

extensible_noattr = (
    extensible_dn + Suppress(":") + matchingrule + Suppress(":=") + value
)
extensible_noattr.setName("extensible_noattr")


def _p_extensible_noattr(s, l, t):
    return [None] + list(t)


extensible_noattr.setParseAction(_p_extensible_noattr)

extensible = extensible_attr | extensible_noattr
extensible.setName("extensible")


def _p_extensible(s, l, t):
    attr, dn, matchingRule, value = t
    return pureldap.LDAPFilter_extensibleMatch(
        matchingRule=matchingRule, type=attr, matchValue=value, dnAttributes=dn
    )


extensible.setParseAction(_p_extensible)
item = simple ^ present ^ substring ^ extensible
item.setName("item")
item.leaveWhitespace()
not_ = Suppress(Literal("!")) + filter_
not_.setParseAction(lambda s, l, t: pureldap.LDAPFilter_not(t[0]))
not_.setName("not")
filterlist = OneOrMore(filter_)
or_ = Suppress(Literal("|")) + filterlist
or_.setParseAction(lambda s, l, t: pureldap.LDAPFilter_or(t))
or_.setName("or")
and_ = Suppress(Literal("&")) + filterlist
and_.setParseAction(lambda s, l, t: pureldap.LDAPFilter_and(t))
and_.setName("and")
filtercomp = and_ | or_ | not_ | item
filtercomp.setName("filtercomp")
filter_ << (
    Suppress(Literal("(").leaveWhitespace())
    + filtercomp
    + Suppress(Literal(")").leaveWhitespace())
)
filter_.setName("filter")
filtercomp.leaveWhitespace()
filter_.leaveWhitespace()

toplevel = StringStart().leaveWhitespace() + filter_ + StringEnd().leaveWhitespace()
toplevel.leaveWhitespace()
toplevel.setName("toplevel")


def parseFilter(s):
    """
    Converting source string to pureldap.LDAPFilter

    Source string is converted to unicode as pyparsing cannot parse bytes
    objects with the rules declared in this module.
    """
    s = to_unicode(s)
    try:
        x = toplevel.parseString(s)
    except ParseException as e:
        raise InvalidLDAPFilter(e.msg, e.loc, e.line)
    assert len(x) == 1
    return x[0]


maybeSubString_value = Combine(OneOrMore(CharsNotIn("*\\\0") | escaped))

maybeSubString_simple = maybeSubString_value.copy()


def _p_maybeSubString_simple(s, l, t):
    return lambda attr: pureldap.LDAPFilter_equalityMatch(
        attributeDesc=pureldap.LDAPAttributeDescription(attr),
        assertionValue=pureldap.LDAPAssertionValue(t[0]),
    )


maybeSubString_simple.setParseAction(_p_maybeSubString_simple)

maybeSubString_present = Literal("*")


def _p_maybeSubString_present(s, l, t):
    return lambda attr: pureldap.LDAPFilter_present(attr)


maybeSubString_present.setParseAction(_p_maybeSubString_present)

maybeSubString_substring = Optional(initial) + any + Optional(final)


def _p_maybeSubString_substring(s, l, t):
    return lambda attr: pureldap.LDAPFilter_substrings(type=attr, substrings=t)


maybeSubString_substring.setParseAction(_p_maybeSubString_substring)

maybeSubString = (
    maybeSubString_simple ^ maybeSubString_present ^ maybeSubString_substring
)


def parseMaybeSubstring(attrType, s):
    try:
        x = maybeSubString.parseString(s)
    except ParseException as e:
        raise InvalidLDAPFilter(e.msg, e.loc, e.line)
    assert len(x) == 1
    fn = x[0]
    return fn(attrType)


if __name__ == "__main__":
    import sys

    for filt in sys.argv[1:]:
        print(repr(parseFilter(filt)))
        print()
